from __future__ import annotations

import math
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

import numpy as np
from cratermaker._cratermaker import morphology_bindings
from numpy.random import Generator
from numpy.typing import ArrayLike, NDArray
from scipy.integrate import quad
from scipy.optimize import root_scalar
from tqdm import tqdm

from cratermaker.components.crater import Crater, CraterFixed, CraterVariable
from cratermaker.components.morphology import Morphology, MorphologyCrater
from cratermaker.components.surface import LocalSurface, Surface
from cratermaker.constants import FloatLike
from cratermaker.utils.general_utils import format_large_units, parameter

if TYPE_CHECKING:
    from cratermaker.components.surface import LocalSurface


@dataclass(frozen=True, slots=True)
class SimpleMoonCraterFixed(CraterFixed):
    rim_height: float | None = None
    """Original rim height of the crater in meters relative to the reference surface."""
    rim_width: float | None = None
    """Original rim width of the crater in meters."""
    floor_depth: float | None = None
    """Original floor depth of the crater in meters relative to the reference surface."""
    floor_diameter: float | None = None
    """Original floor diameter of the crater in meters."""
    peak_height: float | None = None
    """Original central peak height of the crater in meters relative to the reference surface. None for simple craters."""
    ejrim: float | None = None
    """Original ejecta rim thickness of the crater in meters."""

    @property
    def depth_to_diameter(self) -> float | None:
        """
        The depth to diameter ratio of the crater.

        This is computed from `rim_height`-`floor_depth`
        """
        floor_depth = self.floor_depth
        rim_height = self.rim_height
        if floor_depth is not None and rim_height is not None:
            return (rim_height - floor_depth) / self.diameter
        else:
            return None


class SimpleMoonCrater(MorphologyCrater):
    def __init__(self, crater: Crater | None = None, fixed_cls=SimpleMoonCraterFixed, **kwargs):
        super().__init__(crater=crater, fixed_cls=fixed_cls, **kwargs)
        return

    def __str__(self) -> str:
        base = super().__str__()
        return (
            f"{base}\n"
            f"Rim height: {format_large_units(self.rim_height, quantity='length')}\n"
            f"Rim width: {format_large_units(self.rim_width, quantity='length')}\n"
            f"Floor depth: {format_large_units(self.floor_depth, quantity='length')}\n"
            f"Floor diameter: {format_large_units(self.floor_diameter, quantity='length')}\n"
            f"Central peak height: {format_large_units(self.peak_height, quantity='length') if self.peak_height else 'None'}\n"
            f"Ejecta rim thickness: {format_large_units(self.ejrim, quantity='length')}\n"
        )

    @classmethod
    def maker(
        cls,
        crater: Crater | None = None,
        morphology: Morphology | None = None,
        **kwargs: Any,
    ) -> SimpleMoonCrater:
        """
        Initialize a SimpleMoonCrater object either from an existing Crater object or from parameters.

        This generates a specialized Crater object with morphology parameters.

        Parameters
        ----------
        crater : Crater, optional
            The crater object to be converted into a SimpleMoonCrater. If None, then a new crater is created using the provided parameters.
        morphology : Morphology, optional
            The morphology model to use for generating morphology parameters.
        kwargs : Any
            The keyword arguments provided are passed down to :py:meth:`cratermaker.morphology.MorphologyCrater.maker`.  Refer to its documentation for a detailed description of valid keyword arguments.
        """
        from cratermaker.components.morphology import Morphology

        morphology = Morphology.maker(morphology, **kwargs)
        if crater is None:
            crater = super().maker(morphology=morphology, **kwargs)
        args = {}
        diameter_m = crater.diameter
        diameter_km = diameter_m * 1e-3

        if crater.morphology_type in ["simple", "transitional"]:
            args["rim_height"] = 0.043 * diameter_km**1.014 * 1e3
            args["rim_width"] = 0.257 * diameter_km**1.011 * 1e3
            args["floor_depth"] = -0.224 * diameter_km**1.010 * 1e3
            args["floor_diameter"] = 0.200 * diameter_km**1.143 * 1e3
            args["peak_height"] = None
        elif crater.morphology_type in ["complex", "peakring", "multiring"]:
            args["rim_height"] = 0.236 * diameter_km**0.399 * 1e3
            args["rim_width"] = 0.467 * diameter_km**0.836 * 1e3
            args["floor_depth"] = -1.044 * diameter_km**0.301 * 1e3
            args["floor_diameter"] = min(0.187 * diameter_km**1.249 * 1e3, 0.9 * diameter_m)
            args["peak_height"] = 0.032 * diameter_km**0.900 * 1e3
        else:
            raise ValueError(f"Unknown morphology type: {crater.morphology_type}")

        args["ejrim"] = 0.14 * (diameter_m * 0.5) ** 0.74

        return cls(
            crater=crater,
            morphology=morphology,
            **args,
        )


@Morphology.register("simplemoon")
class SimpleMoon(Morphology):
    """
    An operations class for computing the morphology of a crater and applying it to a surface mesh.
    """

    def __init__(
        self,
        surface: Surface | str | None = None,
        ejecta_truncation: FloatLike | None = None,
        dorays: bool = False,
        rng: Generator | None = None,
        rng_seed: int | None = None,
        rng_state: dict | None = None,
        **kwargs: Any,
    ):
        """
        **Warning:** This object should not be instantiated directly. Instead, use the ``.maker()`` method.

        Parameters
        ----------
        surface : str or Surface, optional
            The name of a Surface object, or an instance of Surface, to be associated the morphology model.
        ejecta_truncation : float, optional
            The relative distance from the rim of the crater to truncate the ejecta blanket, default is None, which will compute a
            truncation distance based on where the ejecta thickness reaches a small value.
        dorays : bool, optional
            A flag to determine if the ray pattern should be used instead of the homogeneous ejecta blanket, default is False.
        rng : numpy.random.Generator | None
            |rng|
        rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
            |rng_seed|
        rng_state : dict, optional
            |rng_state|
        **kwargs : Any
            |kwargs|
        """
        super().__init__(surface=surface, rng=rng, rng_seed=rng_seed, rng_state=rng_state, **kwargs)
        object.__setattr__(self, "_ejecta_truncation", None)
        object.__setattr__(self, "_node", None)
        self.ejecta_truncation = ejecta_truncation
        self.dorays = dorays
        return

    def __str__(self) -> str:
        base = super().__str__()
        if self.ejecta_truncation is not None:
            base += f"\nEjecta Trunction: {self.ejecta_truncation:.2f} * crater.radius"
        else:
            base += "\nEjecta Truncation: Off"
        return f"{base}\nEjecta Rays: {self.dorays}"

    def emplace(self, craters: Crater | list[Crater] | None = None, **kwargs: Any) -> list[SimpleMoonCrater]:
        """
        Convenience method to immediately emplace a crater onto the surface.

        Initializes and uses the queue system behind the scenes.

        Parameters
        ----------
        crater : Crater | list[Crater] | None
            The crater or list of craters to be emplaced. If None, then a crater will be created using the provided parameters in kwargs and emplaced.
        kwargs : Any
            |kwargs|

        Returns
        -------
        list[SimpleMoonCrater]
            The list of SimpleMoonCrater objects that were emplaced.
        """
        if craters is None:
            craters = [SimpleMoonCrater.maker(morphology=self, **kwargs)]
        elif isinstance(craters, (list | tuple)) and len(craters) > 0:
            processed_craters = []
            for c in tqdm(
                craters,
                total=len(craters),
                desc="Preparing craters for emplacement",
                unit="crater",
                position=0,
                leave=False,
            ):
                processed_craters.append(SimpleMoonCrater.maker(c, morphology=self, **kwargs))
            craters = processed_craters
        elif isinstance(craters, Crater):
            craters = [SimpleMoonCrater.maker(craters, morphology=self, **kwargs)]
        else:
            raise ValueError(
                "Invalid input for crater emplacement. Must be a Crater object, a list of Crater objects, or None with additional arguments for Crater.maker()."
            )

        return super().emplace(craters, **kwargs)

    def form_crater(self, crater: Crater | SimpleMoonCrater, **kwargs: Any) -> None:
        """
        Form a crater or list of craters on the surface.

        Parameters
        ----------
        crater : SimpleMoonCrater
            The crater object to be formed.
        kwargs : Any
            |kwargs|
        """
        if not isinstance(crater, SimpleMoonCrater):
            crater = SimpleMoonCrater.maker(crater, morphology=self)

        if crater.emplaceable:
            super().form_crater(crater, **kwargs)
        return

    def form_ejecta(self, crater: Crater | SimpleMoonCrater, **kwargs: Any) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """
        Form the ejecta blanket of the crater by altering the elevation variable of the surface mesh.

        Parameters
        ----------
        crater : Crater
            The crater object whose ejecta is to be formed.
        **kwargs : Any
            |kwargs|

        Returns
        -------
        tuple[NDArray[np.float64], NDArray[np.float64]]
            The computed ejecta thickness and intensity at the face and node elevations.
        """
        if not isinstance(crater, SimpleMoonCrater):
            crater = SimpleMoonCrater.maker(crater, morphology=self)
        ejecta_thickness, ejecta_intensity = super().form_ejecta(crater, **kwargs)
        k_ej = self.ejecta_burial_degradation(ejecta_thickness[: crater.ejecta_region.n_face], ejecta_soften_factor=1.50)
        crater.ejecta_region.apply_diffusion(k_ej)

        k_deg = self.degradation_function(crater.diameter, fe=100) * ejecta_intensity[: crater.ejecta_region.n_face]
        crater.ejecta_region.apply_diffusion(k_deg)
        return ejecta_thickness, ejecta_intensity

    def crater_shape(self, crater: SimpleMoonCrater, region: LocalSurface, **kwargs: Any) -> NDArray[np.float64]:
        """
        Compute the crater shape based on the region view and surface.

        Parameters
        ----------
        crater : SimpleMoonCrater
            The crater object containing the parameters for the crater shape.
        region:  LocalSurface
            The region view of the surface mesh centered at the crater center.
        **kwargs : Any
            |kwargs|

        Returns
        -------
        NDArray[np.float64]
            The computed crater shape at the face and node elevations.
        """
        if not isinstance(crater, SimpleMoonCrater):
            crater = SimpleMoonCrater.maker(crater, morphology=self)
        reference_elevation = region.get_reference_surface(reference_radius=crater.radius)

        # Combine distances and references for nodes and faces
        distance = np.concatenate([region.face_distance, region.node_distance])

        original_elevation = np.concatenate([region.face_elevation, region.node_elevation])

        new_elevation = self.crater_profile(crater, distance, reference_elevation)
        elevation_change = new_elevation - original_elevation

        return elevation_change

    def crater_profile(self, crater: SimpleMoonCrater, r: ArrayLike, r_ref: ArrayLike | None = None) -> NDArray[np.float64]:
        """
        Compute the crater profile elevation at a given radial distance.

        Parameters
        ----------
        crater : SimpleMoonCrater
            The crater object containing the parameters for the crater profile.
        r : ArrayLike
            Radial distances from the crater center (in meters).
        r_ref : ArrayLike, optional
            Reference elevation values to be modified by the crater profile.

        Returns
        -------
        elevation : NDArray[np.float64]
            The computed crater elevation profile at each radial point.

        Notes
        -----
        This is a wrapper for a compiled Rust function.
        """
        if not isinstance(crater, SimpleMoonCrater):
            crater = SimpleMoonCrater.maker(crater, morphology=self)
        if r_ref is None:
            r_ref = np.zeros_like(r)

        if np.isscalar(r):
            r = np.array([r], dtype=np.float64)
        elif isinstance(r, (list | tuple)):
            r = np.array(r, dtype=np.float64)

        # flatten r to 1D array
        rflat = np.ravel(r)
        r_ref_flat = np.ravel(r_ref)
        elevation = morphology_bindings.crater_profile(
            rflat,
            r_ref_flat,
            crater.diameter,
            crater.floor_depth,
            crater.floor_diameter,
            crater.rim_height,
            crater.ejrim,
        )
        # reshape elevation to match the shape of r
        elevation = np.array(elevation, dtype=np.float64)
        elevation = np.reshape(elevation, r.shape)

        return elevation

    def ejecta_shape(
        self, crater: SimpleMoonCrater, region: LocalSurface, **kwargs: Any
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """
        Compute the ejecta shape based on the region view and surface.

        Parameters
        ----------
        region : LocalSurface
            The region view of the surface mesh centered at the crat er center.
        crater : SimpleMoonCrater
            The crater object containing the parameters for the ejecta shape.
        **kwargs : Any
            |kwargs|

        Returns
        -------
        NDArray[np.float64]
            The computed ejecta shape at the face and node elevations
        NDArray[np.float64]
            The computed ejecta intensity at the face and node elevations which is used to compute the degradation function. When `dorays` is True, this is the ray intensity function. When `dorays` is False, this is just a constant 1 over the ejecta region.
        """
        if not isinstance(crater, SimpleMoonCrater):
            crater = SimpleMoonCrater.maker(crater, morphology=self)

        distance = np.concatenate([region.face_distance, region.node_distance])
        if self.dorays:
            bearing = np.concatenate([region.face_bearing, region.node_bearing])
            thickness, intensity = self.ejecta_distribution(crater, distance, bearing)
        else:
            thickness = self.ejecta_profile(crater, distance)
            intensity = np.ones_like(thickness)

        return thickness, intensity

    def ejecta_profile(self, crater: SimpleMoonCrater, r: ArrayLike) -> NDArray[np.float64]:
        """
        Compute the ejecta elevation profile at a given radial distance.

        Parameters
        ----------
        crater : SimpleMoonCrater
            The crater object containing the parameters for the ejecta profile.
        r : ArrayLike
            Radial distances from the crater center (in meters).

        Returns
        -------
        elevation : NDArray[np.float64]
            The computed ejecta profile at each radial point.

        Notes
        -----
        This is a wrapper for a compiled Rust function.
        """
        if not isinstance(crater, SimpleMoonCrater):
            crater = SimpleMoonCrater.maker(crater, morphology=self)
        if np.isscalar(r):
            r = np.array([r], dtype=np.float64)
        elif isinstance(r, (list | tuple)):
            r = np.array(r, dtype=np.float64)
        # flatten r to 1D array
        rflat = np.ravel(r)
        elevation = morphology_bindings.ejecta_profile(rflat, crater.diameter, crater.ejrim)
        elevation = np.array(elevation, dtype=np.float64)
        # reshape elevation to match the shape of r
        elevation = np.reshape(elevation, r.shape)
        return elevation

    def ejecta_distribution(self, crater, r: ArrayLike, theta: ArrayLike) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """
        Compute the ejecta thickness distribution modulated by ray patterns.

        Parameters
        ----------
        crater : SimpleMoonCrater
            The crater object containing the parameters for the ejecta distribution.
        r : ArrayLike
            Radial distances from the crater center (in meters).
        theta : ArrayLike
            Angular bearings from the crater center (in degrees).

        Returns
        -------
        thickness : NDArray[np.float64]
            The computed ejecta thickness for each (r, theta) pair.
        ray_intensity : NDArray[np.float64]
            The computed ray intensity for each (r, theta) pair.

        Notes
        -----
        This is a wrapper for a compiled Rust function.
        """
        if not isinstance(crater, SimpleMoonCrater):
            crater = SimpleMoonCrater.maker(crater, morphology=self)
        # flatten r and theta to 1D arrays
        thickness = self.ejecta_profile(crater, r)
        rflat = np.ravel(r)
        theta_flat = np.ravel(theta)
        intensity = morphology_bindings.ray_intensity(
            radial_distance=rflat,
            initial_bearing=np.radians(theta_flat),
            crater_diameter=crater.diameter,
            seed=self.rng.integers(0, 2**32 - 1),
        )
        thickness = np.array(thickness, dtype=np.float64)
        intensity = np.array(intensity, dtype=np.float64)
        thickness *= intensity
        # reshape thickness to match the shape of r and theta
        return np.reshape(thickness, r.shape), np.reshape(intensity, r.shape)

    def ray_intensity(self, crater: SimpleMoonCrater, r: ArrayLike, theta: ArrayLike) -> NDArray[np.float64]:
        """
        Compute the ray pattern intensity modulation at each (r, theta) pair.

        Parameters
        ----------
        crater : SimpleMoonCrater
            The crater object containing the parameters for the ray intensity.
        r : ArrayLike
            Radial distances from the crater center (in meters).
        theta : ArrayLike
            Angular bearings from the crater center (in degrees).

        Returns
        -------
        intensity : NDArray[np.float64]
            The computed ray intensity values.

        Notes
        -----
        This is a wrapper for a compiled Rust function.
        """
        if not isinstance(crater, SimpleMoonCrater):
            crater = SimpleMoonCrater.maker(crater, morphology=self)
        # flatten r and theta to 1D arrays
        rflat = np.ravel(r)
        theta_flat = np.radians(np.ravel(theta))
        intensity = morphology_bindings.ray_intensity(
            rflat,
            theta_flat,
            crater.diameter,
            seed=self.rng.integers(0, 2**32 - 1),
        )
        intensity = np.array(intensity, dtype=np.float64)
        # reshape intensity to match the shape of r and theta
        intensity = np.reshape(intensity, r.shape)
        return intensity

    def ejecta_burial_degradation(self, ejecta_thickness, ejecta_soften_factor=1.50) -> NDArray[np.float64]:
        """
        Computes the change in degradation state due to ejecta burial.

        This function implements a combination of the model by Minton et al. (2019) [#]_.

        Parameters
        ----------
        region : LocalSurface
            The region view of the surface mesh centered at the crater center.
        ejecta_thickness : NDArray[np.float64]
            The computed ejecta thickness at the face and node elevations.

        Returns
        -------
        NDArray[np.float64]
            The computed change in degradation state for all faces in the

        References
        ----------
        .. [#] Minton, D.A., Fassett, C.I., Hirabayashi, M., Howl, B.A., Richardson, J.E., (2019). The equilibrium size-frequency distribution of small craters reveals the effects of distal ejecta on lunar landscape morphology. Icarus 326, 63-87. https://doi.org/10.1016/j.icarus.2019.02.021
        """
        return ejecta_soften_factor * ejecta_thickness**2

    def compute_subpixel_degradation(
        self,
        time_start: float,
        time_end: float,
        **kwargs,
    ) -> None:
        """
        Performs the subpixel degradation.

        This models the combined degradation of the part of the production population that is below the resolution of the mesh on each face. It is called between batches of craters by the `emplace` method.

        Parameters
        ----------
        time_start : float
            The time of the surface at the start of the degradation.
        time_end : float
            The time of the surface at the end of the degradation.
            |kwargs|
        """
        dc_min = 1e-8  # Minimum crater size for subpixel degradation calculation.

        if time_end >= time_start:
            raise ValueError("time_end must be less than time_start.")
        if self.production is None:
            raise RuntimeError("Production model must be set in the Morphology object to compute subpixel degradation.")

        if not hasattr(self, "_Kdiff"):
            self._Kdiff = np.zeros_like(self.surface.face_elevation)

        def _subpixel_degradation(diameter):
            fe = 100.0
            k = self.degradation_function(diameter, fe)
            n = self.production.function(
                diameter=diameter,
                time_start=time_start,
                time_end=time_end,
                validate_inputs=False,
            ).item()
            degradation_region_area = np.pi * (diameter / 2) * fe
            return k * n * degradation_region_area

        for face_indices, dc_max in zip(self.surface.face_bin_indices, self.surface.face_bin_max_sizes, strict=False):
            delta_kdiff, _ = quad(_subpixel_degradation, dc_min, dc_max)
            self._Kdiff[face_indices] += delta_kdiff

        # If any Kdiff values reaches a threshold where a meaningful amount of diffusion will occur on the surface, then go ahead and apply it.
        # Otherwise, degradation will continue to accumulate until the next batch of craters is processed.
        if np.any(self._Kdiff / self.surface.face_area > 1):
            self.apply_subpixel_degradation()
            if self.do_counting:
                measure_rim = kwargs.pop("measure_rim", False)
                self.counting.tally(measure_rim=measure_rim, **kwargs)

        return

    def apply_subpixel_degradation(self, **kwargs) -> None:
        """
        Apply subpixel degradation to the surface using the current Kdiff values.

        This method is called after all craters have been processed and is used to
        apply the accumulated degradation effects.
        """
        if not hasattr(self, "_Kdiff"):
            return

        self.surface.apply_diffusion(self._Kdiff)
        self._Kdiff = np.zeros_like(self.surface.face_elevation)
        return

    def rmax(
        self,
        crater: Crater,
        minimum_thickness: FloatLike,
        feature: str = "ejecta",
    ) -> float:
        """
        Compute the maximum extent of the crater based on the minimum thickness of a feature, or the ejecta_truncation factor, whichever is smaller.

        Parameters
        ----------
        crater : Crater
            The crater object to be used.
        minimum_thickness : FloatLike
            The minimum thickness of the feature blanket in meters.
        feature : str, optional, default = "ejecta"
            The feature to compute the maximum extent. Either "crater" or "ejecta". If "crater" is chosen, the rmax is based
            on where the raised rim is smaller than minimum thickness.

        Returns
        -------
        float
            The maximum extent of the crater or ejecta blanket in meters.
        """
        if not isinstance(crater, SimpleMoonCrater):
            crater = SimpleMoonCrater.maker(crater, morphology=self)

        def _profile_invert_ejecta(r):
            ans = self.ejecta_profile(crater, r) - minimum_thickness
            return ans[0]

        def _profile_invert_crater(r):
            ans = self.crater_profile(crater, r, np.zeros(1)) - minimum_thickness
            return ans[0]

        if feature == "ejecta":
            _profile_invert = _profile_invert_ejecta
        elif feature == "crater":
            _profile_invert = _profile_invert_crater
        else:
            raise ValueError("Unknown feature type. Choose either 'crater' or 'ejecta'")

        # Get the maximum extent
        lower_limit = crater.radius * 1.0001
        upper_limit = self.ejecta_truncation * crater.radius if self.ejecta_truncation else np.pi * self.surface.target.radius

        if _profile_invert(lower_limit) < 0:
            ans = lower_limit
        elif _profile_invert(upper_limit) > 0:
            ans = upper_limit
        else:
            sol = root_scalar(
                _profile_invert,
                bracket=[lower_limit, upper_limit],
                method="brentq",
            )
            ans = sol.root if sol.converged else crater.radius

        return float(ans)

    def degradation_function(
        self,
        diameter: FloatLike,
        fe: FloatLike = 100.0,
    ) -> float:
        """
        Computes the degradation function, which defines the topographic degradation that each crater contributes to the surface.

        This function implements a combination of the model by Minton et al. (2019) [#]_ for small craters and Riedel et al. (2020) [#]_ for large craters. It is currently not well-constrained, so may change in the future.

        Parameters
        ----------
        diameter : FloatLike
            The final diameter of the crater in meters.
        fe : FloatLike, optional
            The degradation function size factor, which is a scaling factor for the degradation function. Default is 100.0.

        Returns
        -------
        float
            The computed degradation function


        References
        ----------
        .. [#] Minton, D.A., Fassett, C.I., Hirabayashi, M., Howl, B.A., Richardson, J.E., (2019). The equilibrium size-frequency distribution of small craters reveals the effects of distal ejecta on lunar landscape morphology. Icarus 326, 63-87. https://doi.org/10.1016/j.icarus.2019.02.021
        .. [#] Riedel, C., Minton, D.A., Michael, G., Orgel, C., Bogert, C.H. van der, Hiesinger, H., 2020. Degradation of Small Simple and Large Complex Lunar Craters: Not a Simple Scale Dependence. Journal of Geophysical Research: Planets 125, e2019JE006273. https://doi.org/10.1029/2019JE006273
        """

        def _kdmare(r, fe, psi):
            """
            The mare-scale degradation function from Minton et al. (2019). See eq. (32).
            """
            kv1 = 0.17
            neq1 = 0.0084
            eta = 3.2
            gamma = 2.0
            beta = 2.0
            kd1 = kv1 * (math.pi * fe**2 * neq1 * (gamma * beta / ((eta - 2.0) * (beta + gamma - eta)))) ** (-gamma / (eta - beta))
            psi = gamma * ((eta - 2.0) / (eta - beta))
            return kd1 * r**psi

        def _smooth_broken(x, a, x_break, alpha_1, alpha_2, delta):
            return a * (x / x_break) ** (alpha_1) * (0.5 * (1.0 + (x / x_break) ** (1.0 / delta))) ** ((alpha_2 - alpha_1) / delta)

        def _kd(r, fe):
            psi_1 = 2.0  # Mare scale power law exponent
            psi_2 = 1.2  # Highlands scale power law exponent
            rb = 0.5e3  # breakpoint radius
            delta = 1.0e0  # Smoothing function

            kd1 = _kdmare(rb, fe, psi_1) / (1 + (psi_1 - psi_2) / psi_1) ** 2
            return _smooth_broken(r, kd1, rb, psi_1, psi_2, delta)

        return float(_kd(diameter / 2, fe))

    @staticmethod
    def overlap_function(crater: SimpleMoonCrater) -> tuple[set[int], set[int]]:
        """
        Get the affected node and face indices for the crater.

        Parameters
        ----------
        crater : SimpleMoonCrater
            The crater object to be used.

        Returns
        -------
        tuple[set[int], set[int]]
            The affected node and face indices for the crater.
        """
        if not crater.emplaceable or crater.affected_node_indices is None or crater.affected_face_indices is None:
            return set(), set()
        else:
            return crater.affected_node_indices, crater.affected_face_indices

    @parameter
    def ejecta_truncation(self) -> float:
        """
        The radius at which the crater is truncated relative to the crater radius.

        Returns
        -------
        float or None
        """
        return self._ejecta_truncation

    @ejecta_truncation.setter
    def ejecta_truncation(self, value: FloatLike | None):
        if value is not None:
            if not isinstance(value, FloatLike):
                raise TypeError("truction_radius must be of type FloatLike")
            self._ejecta_truncation = float(value)
        else:
            self._ejecta_truncation = None

    @parameter
    def dorays(self) -> bool:
        """
        A flag to determine if the ray pattern should be used instead of the homogeneous ejecta blanket.

        Returns
        -------
        bool
        """
        return self._dorays

    @dorays.setter
    def dorays(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise TypeError("dorays must be of type bool")
        self._dorays = value

    class Crater(SimpleMoonCrater):
        def __init__(self, crater: Crater | None = None, **kwargs):
            kwargs["morphology"] = self
            super().__init__(crater=crater, **kwargs)
            return



import math
from pathlib import Path
from typing import Any

from numpy.typing import NDArray

from cratermaker.components.surface import Surface
from cratermaker.components.target import Target
from cratermaker.utils.general_utils import format_large_units, parameter


@Surface.register("icosphere")
class IcosphereSurface(Surface):
    """
    Create a uniform grid configuration using an icosphere. This is the most accurate and efficient way to create a uniform grid, but is limited to a few resolutions.

    Parameters
    ----------
    gridlevel : float, default 8
        The subdivision level of the icosphere.
    target : Target, optional
        The target body or name of a known target body for the impact simulation.
    reset : bool, optional
        Flag to indicate whether to reset the surface. Default is True.
    regrid : bool, optional
        Flag to indicate whether to regrid the surface. Default is False.
    simdir : str | Path
        |simdir|

    Returns
    -------
    IcosphereSurface
        An instance of the IcosphereSurface class initialized with the given pixel size.

    Notes
    -----
    The number of faces of the icosphere is given by the formula:

    .. math::

        n_{face} = 10 * 4^{gridlevel} + 2

    The number of nodes is given by the formula:

    .. math::

        n_{node} = 20 * 4^{gridlevel}
    """

    def __init__(
        self,
        gridlevel: int = 8,
        target: Target | str | None = None,
        reset: bool = True,
        regrid: bool = False,
        simdir: str | Path | None = None,
        **kwargs: Any,
    ):
        super().__init__(target=target, simdir=simdir, **kwargs)
        self.gridlevel = gridlevel
        self._load_from_files(reset=reset, regrid=regrid, **kwargs)

    def __str__(self) -> str:
        base = super().__str__()
        pix_mean = format_large_units(self.pix_mean, quantity="length")
        pix_std = format_large_units(self.pix_std, quantity="length")
        return f"{base}\nGrid Level: {self.gridlevel}\nEffective pixel size: {pix_mean} +/- {pix_std}"

    @property
    def _hashvars(self):
        """
        The variables used to generate the hash.
        """
        return super()._hashvars + [self.gridlevel]

    def _generate_face_distribution(self, **kwargs: Any) -> NDArray:
        """
        Creates the points that define the mesh centers.

        Returns
        -------
        (3,n) ndarray of np.float64
            Array of points on a unit sphere.
        """
        from trimesh.creation import icosphere

        print(f"Generating a mesh with icosphere level {self.gridlevel}.")
        mesh = icosphere(self.gridlevel)
        points = mesh.vertices.T
        return points

    @parameter
    def gridlevel(self) -> int:
        """The subdivision level of the icosphere."""
        return self._gridlevel

    @gridlevel.setter
    def gridlevel(self, value: int) -> None:
        if value < 0:
            raise ValueError("The value of gridlevel must be a non-negative integer.")
        self._gridlevel = int(value)

    @property
    def pix(self) -> float:
        """The approximate face size for a cell of the mesh in meters."""
        if self._pix is None:
            nfaces = 10 * 4**self.gridlevel + 2
            self._pix = (4 * math.pi * self.radius**2 / nfaces) ** 0.5
        return self._pix



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np
from cratermaker._cratermaker import counting_bindings

from cratermaker.components.crater import Crater, CraterFixed, CraterVariable
from cratermaker.constants import PairOfFloats
from cratermaker.utils.general_utils import format_large_units

if TYPE_CHECKING:
    from cratermaker.components.morphology import Morphology
    from cratermaker.components.surface import LocalSurface

# The factor by which the crater tagging region is extended beyond the final rim.
_RIM_BUFFER_FACTOR = 1.5


from abc import abstractmethod
from collections.abc import Callable
from math import pi
from typing import TYPE_CHECKING

from numpy.typing import NDArray
from tqdm import tqdm

from cratermaker.constants import FloatLike
from cratermaker.core.base import ComponentBase, import_components
from cratermaker.utils.general_utils import parameter

if TYPE_CHECKING:
    from cratermaker.components.counting import Counting
    from cratermaker.components.production import Production
    from cratermaker.components.surface import LocalSurface, Surface


class MorphologyCraterVariable(CraterVariable):
    def __init__(self, morphology: Morphology | None = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        object.__setattr__(self, "_morphology", morphology)
        object.__setattr__(self, "_ejecta_region", None)
        object.__setattr__(self, "_crater_region", None)
        object.__setattr__(self, "_face_index", None)
        object.__setattr__(self, "_affected_face_indices", None)
        object.__setattr__(self, "_affected_node_indices", None)
        object.__setattr__(self, "_ejecta_rmax", None)
        object.__setattr__(self, "_emplaceable", None)
        object.__setattr__(self, "_Crater", None)
        return

    def as_dict(self) -> dict:
        """
        Return a dictionary representation of the crater variable properties.
        """
        dict_repr = super().as_dict()
        keys = (
            "face_index",
            "ejecta_rmax",
            "emplaceable",
            "affected_face_indices",
            "affected_node_indices",
            "morphology",
            "ejecta_region",
            "crater_region",
            "emplaceable",
        )
        for key in keys:
            dict_repr[key] = getattr(self, key)

        return dict_repr

    @property
    def _has_initialized_surface_data(self) -> bool:
        """
        A helper property to check if the morphology's surface data has been initialized, which is necessary for many of the variable properties to be computed.

        This is used to avoid unnecessary computations and potential errors when accessing properties that depend on surface data before it is available.
        """
        return self.morphology is not None and self.morphology.surface.uxds is not None

    @property
    def morphology(self) -> Morphology:
        """The morphology model associated with this crater variable object."""
        return self._morphology

    @property
    def emplaceable(self) -> bool | None:
        """Whether this crater is large enough to be emplaced on the surface mesh, which is determined based on whether the crater region could be successfully extracted."""
        return self._emplaceable

    @property
    def face_index(self) -> int | None:
        """The index of the face on the surface mesh where the crater is centered."""
        return self._face_index

    @property
    def ejecta_rmax(self) -> float | None:
        """The maximum radius of the ejecta region for this crater, as determined by the morphology model and is based on the distance that the ejecta thickness falls below the value of the `smallest_length` attribute of the morphology's associated Surface object."""
        return self._ejecta_rmax

    @property
    def affected_face_indices(self) -> set[int] | None:
        """The set of face indices on the surface mesh that are affected by this crater, which is determined based on the crater region and is used by the morphology's queue manager to determine if craters overlap and can be emplaced simultaneously."""
        return self._affected_face_indices

    @property
    def affected_node_indices(self) -> set[int] | None:
        """The set of node indices on the surface mesh that are affected by this crater, which is determined based on the crater region and is used by the morphology's queue manager to determine if craters overlap and can be emplaced simultaneously."""
        return self._affected_node_indices

    @property
    def ejecta_region(self) -> LocalSurface | None:
        """The LocalSurface view extracted around the crater center that encompasses the ejecta blanket, as determined by the morphology model."""
        return self._ejecta_region

    @property
    def crater_region(self) -> LocalSurface | None:
        """The LocalSurface view extracted around the crater center that encompasses a buffered region around the cratered rim, as determined by the morphology model."""
        return self._crater_region


class MorphologyCrater(Crater):
    def __init__(self, crater: Crater | None = None, fixed_cls=CraterFixed, var_cls=MorphologyCraterVariable, **kwargs):
        super().__init__(crater=crater, fixed_cls=fixed_cls, var_cls=var_cls, **kwargs)
        return

    def __str__(self) -> str:
        output = super().__str__()
        output += f"Ejecta region maximum radius: {format_large_units(self.ejecta_rmax, quantity='length')}\n"
        output += f"\nLarge enough to be emplaced on the grid: {self.emplaceable}\n"
        if self.emplaceable:
            output += f"Face index of crater center: {self.face_index}\n"
            output += f"Crater region: {self.crater_region}\n"
            output += f"Ejecta region: {self.ejecta_region}\n"
        return output

    @classmethod
    def maker(
        cls,
        crater: Crater | None = None,
        morphology: Morphology | None = None,
        location: PairOfFloats | None = None,
        relative_location: dict | None = None,
        check_redundant_inputs: bool = True,
        **kwargs: Any,
    ) -> MorphologyCrater:
        """
        Initialize a MorphologyCrater object either from an existing Crater object or from parameters.

        This generates a specialized Crater object with morphology parameters.

        Parameters
        ----------
        crater : Crater, optional
            The crater object to be converted into a SimpleMoonCrater. If None, then a new crater is created using the provided parameters.
        morphology : Morphology, optional
            The morphology model to use for generating morphology parameters.
        location : pair of floats, optional
            The (longitude, latitude) location of the crater.
        relative_location : dict
            Set the crater's location based on the distance and bearing from a reference location. The dictionary is required to have the keys: "distance" and "bearing" with values of distance (in meters) and bearing (in degrees) from the reference point. A third key, "reference_location" should be in the form of a tuple of floats representing the longitude and latitude pair that defines the reference location. The "reference_location" is optional if the morphology.surface contains a `local_location` attribute (such as in the case of a HiResLocal surface type and its derivatives) which will be used if the key is missing or set to None, otherwise it is required. Both location and relative_location cannot be set unless check_redundant_inputs is set to False.
        check_redundant_inputs : bool, optional
            If True, check for redundant inputs such as providing both diameter and radius. Default is True.
        kwargs : Any
            The keyword arguments provided are passed down to :py:meth:`MorphologyCrater.maker() <cratermaker.components.morphology.MorphologyCrater.maker>`.  Refer to its documentation for a detailed description of valid keyword arguments.
        """
        from cratermaker.components.morphology import Morphology

        morphology = Morphology.maker(morphology, **kwargs)
        if relative_location is not None:
            if location is not None and check_redundant_inputs:
                raise ValueError("Cannot provide both location and relative_location")
            if not isinstance(relative_location, dict) or "distance" not in relative_location or "bearing" not in relative_location:
                raise TypeError("relative_location must be a dict with keys 'distance' and 'bearing'.")
            location = morphology.surface.compute_location_from_distance_bearing(**relative_location)

        if crater is None:
            crater = super().maker(location=location, check_redundant_inputs=check_redundant_inputs, **kwargs)
        args = {}

        return cls(
            crater=crater,
            morphology=morphology,
            **args,
        )

    def as_dict(self, ignore_keys: list[str] | tuple[str] = (), skip_complex_data: bool = False, **kwargs) -> dict:
        """
        Return a dictionary representation of the crater properties, including morphology-specific variable properties.

        Parameters
        ----------
        ignore_keys : list[str] or tuple[str], optional
            A list or tuple of property names to ignore when creating the dictionary representation. Default is an empty tuple unless `skip_complex_data` is True, in which case it will be extended to include ("morphology", "affected_face_indices", "affected_node_indices", "ejecta_region", "ejecta_region", "crater_region").
        skip_complex_data : bool, optional
            If True, skip complex data types when creating the dictionary representation. This is useful when serializing the object for saving to a file, as it removes complex data types that may not be serializable. Default is False.
        """
        if skip_complex_data:
            ignore_keys += (
                "morphology",
                "affected_face_indices",
                "affected_node_indices",
                "ejecta_region",
                "crater_region",
            )
        dict_repr = super().as_dict(ignore_keys=ignore_keys, skip_complex_data=skip_complex_data, **kwargs)
        return dict_repr

    def remove_complex_data(self) -> None:
        """
        Remove complex data types from the crater variable properties.

        This is useful for storing a lightweight representation, as it removes complex data types that can be recomputed from the fixed properties and the morphology model when needed. The morphology model is not removed, so that the complex data can be recomputed if needed.
        """
        self._var._affected_face_indices = None
        self._var._affected_node_indices = None
        self._var._ejecta_region = None
        self._var._crater_region = None
        return

    @property
    def face_index(self) -> int | None:
        """The index of the face on the surface mesh where the crater is centered."""
        if self._var._face_index is None and self._has_initialized_surface_data:
            self._var._face_index = self.morphology.surface.find_nearest_face(self.location)
        return self._var._face_index

    @property
    def ejecta_rmax(self) -> float | None:
        """The maximum radius of the ejecta region for this crater, as determined by the morphology model and is based on the distance that the ejecta thickness falls below the value of the `smallest_length` attribute of the morphology's associated Surface object."""
        if self._var._ejecta_rmax is None and self._has_initialized_surface_data:
            self._var._ejecta_rmax = self.morphology.rmax(
                self, minimum_thickness=self.morphology.surface.smallest_length, feature="ejecta"
            )
        return self._var._ejecta_rmax

    @property
    def ejecta_region(self) -> LocalSurface | None:
        """
        The LocalSurface view extracted around the crater center that encompasses the ejecta blanket, as determined by the morphology model.

        This is extracted from the morphology's associated Surface object based on the location of the crater and the ejecta_rmax distance. If the ejecta region cannot be extracted (e.g. if it is smaller than a single face of the mesh) this property will be set to None and the crater will be marked as not emplaceable.
        """
        if self._var._ejecta_region is None and self.crater_region is not None:
            self._var._ejecta_region = self.morphology.surface.extract_region(
                location=self.location,
                region_radius=self.ejecta_rmax,
            )
        return self._var._ejecta_region

    @property
    def crater_region(self) -> LocalSurface | None:
        """
        The LocalSurface view extracted around the crater center that encompasses a buffered region around the cratered rim.

        This is extracted from the morphology's associated Surface object based on the location of the crater and a radius that extends by a `_RIM_BUFFER_FACTOR` constant times the crater radius. If the crater region cannot be extracted (e.g. if it is smaller than a single face of the mesh) this property will be set to None and the crater will be marked as not emplaceable.
        """
        if self._var._crater_region is None and self._has_initialized_surface_data:
            self._var._crater_region = self.morphology.surface.extract_region(
                location=self.location,
                region_radius=_RIM_BUFFER_FACTOR * self.radius,
            )
        return self._var._crater_region

    @property
    def affected_face_indices(self) -> set[int] | None:
        """
        The set of face indices on the surface mesh that are affected by this crater, which is determined based on the crater region.

        This is used by the morphology's queue manager to determine if craters overlap and can be emplaced simultaneously.
        """
        if self._var._affected_face_indices is None and self._has_initialized_surface_data and self.emplaceable:
            if self.ejecta_region is not None:
                if isinstance(self.ejecta_region.face_indices, slice):
                    self._var._affected_face_indices = set(
                        np.arange(self.morphology.surface.n_face)[self.ejecta_region.face_indices]
                    )
                else:
                    self._var._affected_face_indices = set(self.ejecta_region.face_indices)
            else:
                self._var._affected_face_indices = set()

        return self._var._affected_face_indices

    @property
    def affected_node_indices(self) -> set[int] | None:
        """
        The set of node indices on the surface mesh that are affected by this crater, which is determined based on the crater region.

        This is used by the morphology's queue manager to determine if craters overlap and can be emplaced simultaneously.
        """
        if self._var._affected_node_indices is None and self._has_initialized_surface_data and self.emplaceable:
            if self.ejecta_region is not None:
                if isinstance(self.ejecta_region.node_indices, slice):
                    self._var._affected_node_indices = set(
                        np.arange(self.morphology.surface.n_node)[self.ejecta_region.node_indices]
                    )
                else:
                    self._var._affected_node_indices = set(self.ejecta_region.node_indices)
            else:
                self._var._affected_node_indices = set()

        return self._var._affected_node_indices

    @property
    def emplaceable(self) -> bool | None:
        """Whether this crater is large enough to be emplaced on the surface mesh, which is determined based on whether the crater region could be successfully extracted."""
        if self._var._emplaceable is None and self._has_initialized_surface_data:
            self._var._emplaceable = self.crater_region is not None
        return self._var._emplaceable

    @property
    def measured_rim_height(self) -> float | None:
        """The measured rim height of the crater, which is determined based on the morphology model's crater shape and the surface elevation data in the crater region."""
        if self.crater_region is not None:
            self.crater_region.compute_desloped_face_elevation()
            self._var._measured_rim_height = counting_bindings.measure_rim_height(self.crater_region, self)
        return self._var._measured_rim_height

    @property
    def measured_floor_depth(self) -> float | None:
        """The measured floor depth of the crater, which is determined based on the morphology model's crater shape and the surface elevation data in the crater region."""
        if self.crater_region is not None:
            self.crater_region.compute_desloped_face_elevation()
            self._var._measured_floor_depth = counting_bindings.measure_floor_depth(self.crater_region, self)
        return self._var._measured_floor_depth

    @property
    def measured_depth_to_diameter(self) -> float | None:
        """
        The measured depth to diameter ratio of the crater.

        This is computed from `measured_rim_height`-`measured_floor_depth`
        """
        if self.crater_region is not None:
            self.crater_region._desloped_face_elevation = None
            floor_depth = self.measured_floor_depth
            rim_height = self.measured_rim_height
            return (rim_height - floor_depth) / self.measured_diameter
        else:
            return None


class Morphology(ComponentBase):
    _registry: dict[str, Morphology] = {}
    """
    The base class for Morphology models.
    """

    def __init__(
        self,
        surface: Surface | str | None = None,
        production: Production | str | None = None,
        counting: Counting | str | None = None,
        do_subpixel_degradation: bool = True,
        do_slope_collapse: bool = True,
        do_counting: bool | None = None,
        **kwargs: Any,
    ) -> None:
        """
        **Warning:** This object should not be instantiated directly. Instead, use the ``.maker()`` method.

        Parameters
        ----------
        surface : str or Surface, optional
            The name of a Surface object, or an instance of Surface, to be associated the morphology model.
        production : str, Production, optional
            The name of a Production object, or an instance of Production, to be associated with the morphology model. This is used for subpixel degradation in the emplace method. It is otherwise ignored if do_subpixel_degradation is False.
        counting : str, Counting or None, optional
            The name of a Counting object, or an instance of Counting, to be associated with the morphology model. This is used to record crater counts during emplacement. If None, no counting will be performed.
        do_subpixel_degradation : bool, optional
            If True, subpixel degradation will be performed during the emplacement of craters. Default is True.
        do_slope_collapse : bool, optional
            If True, slope collapse will be performed during the emplacement of craters. Default is True.
        do_counting : bool or None, optional
            If True, counting will be performed during emplacement if a Counting object is provided. If False, no counting will be performed even if a Counting object is provided. If None (default), the presence or absence of a Counting object will determine whether counting is performed.
        **kwargs : Any
            |kwargs|

        """
        from cratermaker.components.counting import Counting
        from cratermaker.components.production import Production
        from cratermaker.components.surface import Surface

        super().__init__(**kwargs)
        object.__setattr__(self, "_production", None)
        object.__setattr__(self, "_counting", None)
        object.__setattr__(self, "_do_counting", None)
        object.__setattr__(self, "_excavated_volume", None)
        object.__setattr__(self, "_Crater", None)

        # Because a Surface object is associated with a Counting object, we should first check to see if we are receiving a Counting object first so that we don't end up creating a spurious Surface object that we don't want.
        if surface is None and isinstance(counting, Counting):
            self.surface = counting.surface
            self.counting = counting
        else:
            self.surface = Surface.maker(surface, **kwargs)
            if counting is not None:
                crater_cls = kwargs.pop("Crater", self.Crater)
                self.counting = Counting.maker(counting, surface=self.surface, Crater=crater_cls, **kwargs)

        if self.counting is not None:
            self.counting.morphology = self  # Associated counting and morphology with each other

        if do_counting is not None:
            if do_counting and self.counting is None:
                raise ValueError("do_counting is True but no counting object was provided or created.")
            else:
                self.do_counting = do_counting

        self.do_subpixel_degradation = do_subpixel_degradation
        self.do_slope_collapse = do_slope_collapse
        if do_subpixel_degradation:
            # Be sure to use the associated Target object from the Surface object when initializing a new Production object
            if not isinstance(production, Production):
                kwargs["target"] = self.surface.target
            self.production = Production.maker(production, **kwargs)
        return

    def __str__(self) -> str:
        base = super().__str__()
        str_repr = f"{base}\n"
        str_repr += f"\n{self.counting}\n"
        str_repr += f"Do slope collapse: {self.do_slope_collapse}\n"
        str_repr += f"Do subpixel degradation: {self.do_subpixel_degradation}\n"
        if self.do_subpixel_degradation:
            str_repr += f"\n{self.production}\n"
        return str_repr

    @classmethod
    def maker(
        cls,
        morphology: str | type[Morphology] | Morphology | None = None,
        surface: Surface | str | None = None,
        production: Production | str | None = None,
        counting: Counting | str | None = None,
        do_subpixel_degradation: bool = True,
        do_slope_collapse: bool = True,
        **kwargs: Any,
    ) -> Morphology:
        """
        Initialize a Morphology model with the given name or instance.

        Parameters
        ----------
        morphology : str or Morphology or None
            The name of the morphology model to use, or an instance of Morphology. If None, the default "simplemoon" is used.
        surface : str or Surface, optional
            The name of a Surface object, or an instance of Surface, to be associated the morphology model.
        production : str or Production, optional
            The name of a Production object, or an instance of Production, to be associated with the morphology model. This is used for subpixel degradation in the emplace method. It is otherwise ignored.
        counting : str or Counting, optional
            The name of a Counting object, or an instance of Counting, to be associated with the morphology model. This is used to record crater counts during emplacement. If None, no counting will be performed.
        do_subpixel_degradation : bool, optional
            If True, subpixel degradation will be performed during the emplacement of craters. Default is True.
        do_slope_collapse : bool, optional
            If True, slope collapse will be performed during the emplacement of craters. Default is True.
        **kwargs : Any
            |kwargs|

        Returns
        -------
        Morphology
            An instance of the specified Morphology model (e.g. SimpleMoon).

        Raises
        ------
        KeyError
            If the specified morphology model name is not found in the registry.
        TypeError
            If the specified morphology model is not a string or a subclass of Morphology.
        """
        # Call the base class version of make and pass the morphology argument as the component argument
        if morphology is None:
            morphology = "simplemoon"
        morphology = super().maker(
            component=morphology,
            surface=surface,
            production=production,
            counting=counting,
            do_subpixel_degradation=do_subpixel_degradation,
            do_slope_collapse=do_slope_collapse,
            **kwargs,
        )
        return morphology

    def emplace(self, craters: Crater | list[Crater] | None = None, **kwargs: Any) -> list[Crater]:
        """
        Convenience method to immediately emplace a crater onto the surface.

        Initializes and uses the queue system behind the scenes.

        Parameters
        ----------
        craters : Crater or list of Crater objects, optional
            The Crater object(s) to be emplaced. If provided, this will be used directly. Otherwise, a single crater will be generated based on the keyword arguments.
        **kwargs : Any
            |kwargs|

        Returns
        -------
        list of Crater or None
            The crater(s) that were emplaced. If no craters were provided and a crater could not be generated based on the keyword arguments, an empty list is returned.

        Notes
        -----
        The keyword arguments provided are passed down to :py:meth:`Crater.maker`.  Refer to its documentation for a detailed description of valid keyword arguments.

        Examples
        --------
        .. code-block:: python

            from cratermaker import Morphology, Crater

            morphology = Morphology.maker()

            # Create a crater with specific diameter
            morphology.emplace(diameter=10.0e3)

            # Create a crater based on a projectile with given mass and projectile_velocity
            morphology.emplace(projectile_mass=1e15, projectile_velocity=20e3)

            # Create a crater with a specific transient diameter and location
            morphology.emplace(transient_diameter=50e3, location=(43.43, -86.92))

            # Create multiple craters
            craters = [Crater.maker(diameter=20.0e3), Crater.maker(diameter=20.0e3)]
            morphology.emplace(craters)
        """
        if self._queue_manager is None:
            self._init_queue_manager()

        if craters is None:
            craters = [MorphologyCrater.maker(**kwargs)]
        elif isinstance(craters, MorphologyCrater):
            craters = [craters]
        elif isinstance(craters, Crater):
            craters = [MorphologyCrater.maker(crater=craters, morphology=self)]

        if isinstance(craters, list) and len(craters) > 0:
            for c in craters:
                self._enqueue_crater(c)

        self._process_queue(**kwargs)

        return craters

    def form_crater(self, crater: Crater, **kwargs: Any) -> None:
        """
        Form the interior of the crater by altering the elevation variable of the surface mesh.

        Parameters
        ----------
        crater : Crater
            The crater object to be formed.
        **kwargs : Any
            |kwargs|
        """
        if not isinstance(crater, Crater):
            raise TypeError("crater must be an instance of Crater")
        if not isinstance(crater, MorphologyCrater):
            crater = MorphologyCrater.maker(crater=crater, morphology=self)
        if not crater.emplaceable:
            return

        crater_area = pi * crater.crater_region.radius**2

        # Check to make sure that the face at the crater location is not smaller than the crater area
        if crater_area > self.surface.face_area[crater.face_index]:
            elevation_change = self.crater_shape(crater, crater.ejecta_region)
            crater.ejecta_region.update_elevation(elevation_change)
            if self.do_slope_collapse:
                crater.ejecta_region.slope_collapse()
            self._excavated_volume = crater.ejecta_region.compute_volume(elevation_change[: crater.ejecta_region.n_face])

            # Remove any ejecta from the interior of the crater
            inner_crater_region = crater.crater_region.extract_subregion(crater.radius)
            if inner_crater_region is not None:
                inner_crater_region.add_data(
                    "ejecta_thickness",
                    long_name="ejecta thickness",
                    units="m",
                    data=0.0,
                    overwrite=True,
                )

            # Record the crater to the counting layer
            if self.do_counting:
                self.counting.add(crater, **kwargs)

            self.form_ejecta(crater, **kwargs)
        return

    def form_ejecta(self, crater: Crater, **kwargs: Any) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """
        Form the ejecta blanket of the crater by altering the elevation variable of the surface mesh.

        Parameters
        ----------
        crater : Crater
            The crater object whose ejecta is to be formed.
        **kwargs : Any
            |kwargs|

        Returns
        -------
        tuple[NDArray[np.float64], NDArray[np.float64]]
            The computed ejecta thickness and intensity at the face and node elevations. |
        """
        if not isinstance(crater, Crater):
            raise TypeError("crater must be an instance of Crater")
        if not isinstance(crater, MorphologyCrater):
            crater = MorphologyCrater.maker(crater=crater, morphology=self)

        if not self._excavated_volume:
            return
        ejecta_thickness, ejecta_intensity = self.ejecta_shape(crater, crater.ejecta_region)
        ejecta_volume = crater.ejecta_region.compute_volume(ejecta_thickness[: crater.ejecta_region.n_face])
        conservation_factor = -self._excavated_volume / ejecta_volume
        ejecta_thickness *= conservation_factor

        crater.ejecta_region.add_data(
            "ejecta_thickness",
            long_name="ejecta thickness",
            units="m",
            data=ejecta_thickness[: crater.ejecta_region.n_face],
        )

        crater.ejecta_region.update_elevation(ejecta_thickness)

        return ejecta_thickness, ejecta_intensity

    def _init_queue_manager(self) -> None:
        """
        Initialize the crater queue manager with a surface-dependent overlap function.
        """
        self._queue_manager = self.CraterQueueManager(self.overlap_function)

    def _enqueue_crater(
        self,
        crater: Crater,
        **kwarg: Any,
    ) -> None:
        """
        Add a crater to the queue for later emplacement.

        Automatically initializes the queue manager if it hasn't been set.

        Parameters
        ----------
        crater : Crater
            The crater object to enqueue.
        **kwarg : Any
            |kwargs|

        Raises
        ------
        RuntimeError
            If the queue manager must be initialized but no surface is provided.
        """
        if self._queue_manager is None:
            if self.surface is None:
                raise RuntimeError("Surface must be provided to initialize queue manager.")
            self._init_queue_manager()
        if crater.emplaceable:
            self._queue_manager.push(crater)

    def _process_queue(self, **kwargs) -> None:
        """
        Process all queued craters in the order they were added, forming non-overlapping batches and applying each to the surface.

        Raises
        ------
        RuntimeError
            If the queue manager has not been initialized.
        """
        if not hasattr(self, "_queue_manager"):
            raise RuntimeError("Queue manager has not been initialized. Call _init_queue_manager first.")

        from concurrent.futures import ThreadPoolExecutor, as_completed

        def _batch_process(pbar=None):
            tally_cadence = 2000
            nacumulated = 0
            while not self._queue_manager.is_empty:
                batch = self._queue_manager.peek_next_batch()

                for crater in batch:
                    self.form_crater(crater)
                    if pbar is not None:
                        pbar.update(1)

                if self.do_subpixel_degradation and len(batch) > 1:
                    # If the craters have time values attached to them, we can perform subpixel degradation between time values
                    timevals = [crater.time for crater in batch if crater.time is not None]
                    if len(timevals) > 1:
                        self.compute_subpixel_degradation(time_start=max(timevals), time_end=min(timevals), **kwargs)

                self._queue_manager.pop_batch(batch)
                nacumulated += len(batch)
                if self.do_counting and nacumulated >= tally_cadence:
                    measure_rim = kwargs.pop("measure_rim", False)
                    self.counting.tally(measure_rim=measure_rim, quiet=False, **kwargs)
                    nacumulated = 0

            return

        total_craters = len(self._queue_manager._queue)
        if total_craters > 10:
            with tqdm(
                total=total_craters,
                desc="Emplacing craters",
                position=0,
                leave=False,
                unit="craters",
                smoothing=10 / total_craters,
            ) as pbar:
                _batch_process(pbar)
        else:
            _batch_process()

        if self.do_subpixel_degradation:
            self.apply_subpixel_degradation(**kwargs)
        if self.do_counting:
            self.counting.tally(**kwargs)
        return

    @abstractmethod
    def crater_shape(
        self,
        crater: Crater,
        region: LocalSurface,
        **kwarg: Any,
    ) -> NDArray[np.float64]: ...

    @abstractmethod
    def ejecta_shape(
        self,
        crater: Crater,
        region: LocalSurface,
        **kwarg: Any,
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]: ...

    @abstractmethod
    def rmax(
        self,
        crater: Crater,
        minimum_thickness: FloatLike,
        feature: str = "ejecta",
    ) -> float: ...

    @staticmethod
    @abstractmethod
    def overlap_function(
        crater: Crater,
    ) -> tuple[set[int], set[int]]: ...

    @property
    def surface(self) -> Surface:
        """The Surface object associated with this morphology model."""
        return self._surface

    @surface.setter
    def surface(self, surface: Surface) -> None:
        """Set the Surface object associated with this morphology model."""
        from cratermaker.components.surface import Surface

        if isinstance(surface, Surface):
            self._surface = surface
        elif surface is None or isinstance(surface, str):
            self._surface = Surface.maker(surface)
        else:
            raise TypeError("surface must be an instance of Surface, a string, or None")
        self._queue_manager: self.CraterQueueManager | None = None

    @property
    def production(self) -> Production:
        """The Production object associated with this morphology model."""
        return self._production

    @production.setter
    def production(self, production: Production) -> None:
        from cratermaker.components.production import Production

        if isinstance(production, Production):
            self._production = production
        elif production is None or isinstance(production, str):
            self._production = Production.maker(production)
        else:
            raise TypeError("production must be an instance of Production a string or None")

    @property
    def counting(self) -> Counting:
        """The Counting object associated with this morphology model."""
        return self._counting

    @counting.setter
    def counting(self, counting: Counting) -> None:
        """Set the Counting object associated with this morphology model."""
        from cratermaker.components.counting import Counting

        if isinstance(counting, Counting):
            # Make sure the associated Crater class is correct for this Morphology class
            if not issubclass(counting.Crater, self.Crater):
                counting.Crater = self.Crater
            self._counting = counting
        elif isinstance(counting, str):
            self._counting = Counting.maker(counting, surface=self.surface, Crater=self.Crater)
        elif counting is None:
            self.do_counting = False
            self._counting = None
            return
        else:
            raise TypeError("counting must be an instance of Counting or a string")
        self.do_counting = True

    @parameter
    def do_subpixel_degradation(self) -> bool:
        """Whether to perform subpixel degradation during crater emplacement."""
        return self._do_subpixel_degradation

    @do_subpixel_degradation.setter
    def do_subpixel_degradation(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise TypeError("do_subpixel_degradation must be a boolean value")
        self._do_subpixel_degradation = value

    @parameter
    def do_slope_collapse(self) -> bool:
        """Whether to perform slope collapse during crater emplacement."""
        return self._do_slope_collapse

    @do_slope_collapse.setter
    def do_slope_collapse(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise TypeError("do_slope_collapse must be a boolean value")
        self._do_slope_collapse = value

    @parameter
    def do_counting(self) -> bool:
        """Whether to perform crater counting during crater emplacement."""
        return self._do_counting

    @do_counting.setter
    def do_counting(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise TypeError("do_counting must be a boolean value")
        self._do_counting = value

    class Crater(MorphologyCrater):
        def __init__(self, crater: Crater | None = None, **kwargs):
            kwargs["morphology"] = self
            super().__init__(crater=crater, **kwargs)
            return

    class CraterQueueManager:
        """
        A manager for craters awaiting emplacement. Craters are processed in order (FIFO) but batches of non-overlapping craters can be processed simultaneously.

        Parameters
        ----------
        overlap_fn : Callable[[Crater], tuple[set[int], set[int]]]
            A function that takes a crater and returns a tuple of (node indices, face indices) affected.
        """

        def __init__(self, overlap_fn: Callable[[Crater], tuple[set[int], set[int]]]):
            self._queue: list[Crater] = []
            self._overlap_fn = overlap_fn

        def push(self, crater: Crater) -> None:
            """
            Add a crater to the end of the queue.

            Parameters
            ----------
            crater : Crater
                The crater to be added to the queue.

            Raises
            ------
            TypeError
                If the provided crater is not an instance of Crater.
            """
            if not isinstance(crater, Crater):
                raise TypeError("crater must be an instance of Crater")
            self._queue.append(crater)
            return

        def peek_next_batch(self) -> list[Crater]:
            """
            Return a list of the next batch of craters that do not overlap with each other or the current active region.

            Returns
            -------
            list of Crater
                The next batch of craters that can be processed simultaneously without overlap.
            """
            batch = []
            reserved_nodes = set()
            reserved_faces = set()
            for crater in self._queue:
                node_indices, face_indices = self._overlap_fn(crater)
                if reserved_nodes.isdisjoint(node_indices) and reserved_faces.isdisjoint(face_indices):
                    batch.append(crater)
                    reserved_nodes.update(node_indices)
                    reserved_faces.update(face_indices)
                else:
                    break
            return batch

        def pop_batch(self, batch: list[Crater]) -> None:
            """
            Remove a processed batch of craters from the queue.

            Parameters
            ----------
            batch : list of Crater
                The batch of craters to be removed from the queue.

            Raises
            ------
            ValueError
                If any crater in the batch is not found in the queue.
            """
            for crater in batch:
                self._queue.remove(crater)
                crater.remove_complex_data()
            return

        @property
        def is_empty(self) -> bool:
            """Return True if the queue is empty, False otherwise."""
            return len(self._queue) == 0


import_components(__name__, __path__)



from typing import Any

import numpy as np
from numba import njit
from numpy import cross
from numpy.linalg import norm
from numpy.random import Generator
from numpy.typing import NDArray
from scipy.stats import truncnorm
from uxarray import Grid

from cratermaker.constants import FloatLike
from cratermaker.core.base import _rng_init


def get_random_location(
    size: int = 1,
    rng: Generator | None = None,
    rng_seed: int | None = None,
    rng_state: dict | None = None,
    **kwargs: Any,
) -> NDArray[np.float64]:
    """
    Computes random longitude and latitude values.

    Generates a set of latitude and longitude values that are uniformly distributed on the surface of a sphere.

    Parameters
    ----------
    size : int or tuple of ints, optional
        The number of samples to generate. If size is None (the default), a single tuple is returned. If size is greater than 1,
        then a structured array with fields 'lon' and 'lat' is returned.
    rng : numpy.random.Generator | None
        |rng|
    rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
        |rng_seed|
    rng_state : dict, optional
        |rng_state|
    **kwargs : Any
        |kwargs|

    Returns
    -------
    A structured numpy array with the location data in the format [('lon', 'f8'), ('lat', 'f8')].
    """
    rng, _ = _rng_init(rng=rng, rng_seed=rng_seed, rng_state=rng_state, **kwargs)

    u = rng.uniform(size=size)
    v = rng.uniform(size=size)

    # Compute the angles theta and phi
    theta = 2 * np.pi * u
    phi = np.arccos(2 * v - 1)

    # Convert to lon/lat
    lon = np.rad2deg(theta - np.pi)  # Use the convention that longitude is in the range [-180, 180]
    lat = np.rad2deg(phi - np.pi / 2.0)

    lonlat_arr = np.empty(size, dtype=[("lon", "f8"), ("lat", "f8")])

    # Reshape lat and lon to the original size if necessary
    lon = lon.reshape(size)
    lat = lat.reshape(size)

    # Combine lat and lon into a structured array
    lonlat_arr["lon"] = lon
    lonlat_arr["lat"] = lat

    return lonlat_arr


@njit
def _get_one_random_location(face_nodes, node_x, node_y, node_z, rng_vals):
    valid_nodes = face_nodes[face_nodes >= 0]
    n = len(valid_nodes)
    n_tris = n - 2
    tris = np.empty((n_tris, 3), dtype=np.int64)
    for j in range(n_tris):
        tris[j, 0] = 0
        tris[j, 1] = j + 1
        tris[j, 2] = j + 2

    n_valid = len(valid_nodes)
    vertices = np.empty((n_valid, 3), dtype=np.float64)
    for i in range(n_valid):
        idx = valid_nodes[i]
        vertices[i, 0] = node_x[idx]
        vertices[i, 1] = node_y[idx]
        vertices[i, 2] = node_z[idx]

    areas = np.empty(len(tris))
    for i in range(len(tris)):
        j0, j1, j2 = tris[i]
        v0, v1, v2 = vertices[j0], vertices[j1], vertices[j2]
        areas[i] = 0.5 * norm(cross(v1 - v0, v2 - v0))
    areas /= areas.sum()
    cum_areas = np.cumsum(areas)

    # triangle selection
    tri_idx = np.searchsorted(cum_areas, rng_vals[0])
    j0, j1, j2 = tris[tri_idx]
    v0, v1, v2 = vertices[j0], vertices[j1], vertices[j2]

    r1, r2 = rng_vals[1], rng_vals[2]
    if r1 + r2 > 1.0:
        r1, r2 = 1.0 - r1, 1.0 - r2
    r0 = 1.0 - r1 - r2
    return r0 * v0 + r1 * v1 + r2 * v2


def get_random_location_on_face(
    grid: Grid,
    face_index: int | NDArray[np.int64],
    rng: Generator | None = None,
    rng_seed: int | None = None,
    rng_state: dict | None = None,
    **kwargs: Any,
) -> NDArray[np.float64]:
    """
    Generate a random coordinate within a given face of an unstructured mesh.

    Parameters
    ----------
    grid : uxarray.Grid
        The grid object containing the mesh information.
    face_index : int | NDArray[np.int64]
        The index or array of indices of the face within the grid to obtain the random sample.
    rng : numpy.random.Generator | None
        |rng|
    rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
        |rng_seed|
    rng_state : dict, optional
        |rng_state|
    **kwargs : Any
        |kwargs|

    Returns
    -------
    A structured numpy array with the location data in the format of the same shape as face_index [('lon', 'f8'), ('lat', 'f8')].
    """
    from uxarray.grid import coordinates

    rng, _ = _rng_init(rng=rng, rng_seed=rng_seed, rng_state=rng_state, **kwargs)

    # Extract node indices for the given face(s)
    face_index = np.atleast_1d(face_index)
    size = len(face_index)
    node_indices = grid.face_node_connectivity[face_index, :]

    # Prepare index arrays
    locations = np.empty(size, dtype=[("lon", "f8"), ("lat", "f8")])

    node_x = grid.node_x.values
    node_y = grid.node_y.values
    node_z = grid.node_z.values
    rng_vals = rng.random((size, 3))
    for i in range(size):
        ind = node_indices.values[i]
        ind = ind[ind >= 0]
        p = _get_one_random_location(ind, node_x, node_y, node_z, rng_vals=rng_vals[i, :])
        lon, lat = coordinates._xyz_to_lonlat_deg(*p)
        locations["lon"][i] = np.float64(lon)
        locations["lat"][i] = np.float64(lat)

    return locations


def get_random_impact_angle(
    size: int | tuple[int, ...] = 1,
    rng: Generator | None = None,
    rng_seed: int | None = None,
    rng_state: dict | None = None,
    **kwargs: Any,
) -> NDArray[np.float64]:
    """
    Sample impact angles from a distribution centered on 45deg.

    For the theory, see Shoemaker (1962) "Interpretation of lunar craters."

    Parameters
    ----------
    size : int or tuple of ints, optional
        The number of samples to generate. If the shape is (m, n, k), then m * n * k samples are drawn. If size is None (the default), a single value is returned if `diameters` is a scalar, otherwise an array of samples is returned with the same size as `diameters`.
    rng : numpy.random.Generator | None
        |rng|
    rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
        |rng_seed|
    rng_state : dict, optional
        |rng_state|
    **kwargs : Any
        |kwargs|

    Returns
    -------
    ndarray of np.float64
        An array of impact angles (in degrees).
    """
    rng, _ = _rng_init(rng=rng, rng_seed=rng_seed, rng_state=rng_state, **kwargs)

    u = np.sqrt(rng.uniform(size=size))
    impact_angle = np.arcsin(u)
    return np.rad2deg(impact_angle)


def get_random_impact_direction(
    size: int | tuple[int, ...] = 1,
    rng: Generator | None = None,
    rng_seed: int | None = None,
    rng_state: dict | None = None,
    **kwargs: Any,
) -> NDArray[np.float64]:
    """
    Sample impact direction from a uniform distribution.

    Parameters
    ----------
    size : int or tuple of ints, optional
        The number of samples to generate. If the shape is (m, n, k), then m * n * k samples are drawn. If size is None (the default), a single scalar value is returned.
    rng : numpy.random.Generator | None
        |rng|
    rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
        |rng_seed|
    rng_state : dict, optional
        |rng_state|
    **kwargs : Any
        |kwargs|

    Returns
    -------
    ndarray of np.float64
        An array of impact angles (in degrees).
    """
    rng, _ = _rng_init(rng=rng, rng_seed=rng_seed, rng_state=rng_state, **kwargs)

    pdir = rng.uniform(0.0, 360.0, size=size)
    return pdir


def get_random_size(
    diameters: NDArray[np.float64],
    cdf: NDArray[np.float64],
    size: int | tuple[int, ...] | None = None,
    mu: int | tuple[int, ...] | None = None,
    rng: Generator | None = None,
    rng_seed: int | None = None,
    rng_state: dict | None = None,
    **kwargs: Any,
) -> NDArray[np.float64]:
    """
    Sample diameters from a cumulative size-frequency distribution (SFD).

    Given an array of diameters and optionally a cumulative distribution function (CDF), this function generates new diameter values that follow the specified distribution. The SFD is treated as a continuous function that interpolates between the provided diameters, which are assumed to represent a power-law distribution.

    Parameters
    ----------
    diameters : array_like
        An array of diameters from which the SFD is constructed. Must be 1-dimensional.
    cdf : array_like
        The cumulative distribution function corresponding to `diameters`. Must be the same size as `diameters` and must be monotonically increasing with decreasing diameter value.
    size : int or tuple of ints, optional
        The number of samples to generate. If the shape is (m, n, k), then m * n * k samples are drawn. If size is None and mu is None then a single value is returned. Note: mu and size are mutually exclusive.
    mu : int or tuple of ints, optional
        The expected number of samples to generate using a Poisson random number genertor. If the shape is (m, n, k), then m * n * k samples are drawn. Note: mu and size are mutually exclusive.
    rng : numpy.random.Generator | None
        |rng|
    rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
        |rng_seed|
    rng_state : dict, optional
        |rng_state|
    **kwargs : Any
        |kwargs|

    Returns
    -------
    ndarray of np.float64
        An array of sampled diameter values from the SFD.

    Notes
    -----
    The SFD is assumed to be a continuous distribution that follows a power-law between the provided discrete diameter values. Linear interpolation in log-space is used to sample new values between the known diameters.
    A small amount of random noise (of the order 1e-3 the diameter value) is added to the final diameter value to ensure that diameter values are unlikely to be identical, even when an input CDF could lead to identical diameter values.

    Examples
    --------
    >>> diameters = np.array([100.0, 56.0, 32.0, 18.0, 10.0])
    >>> ncumul = np.array([1.0, 0.51, 0.21, 0.06, 0.01])
    >>> sample_from_sfd(diameters, cdf=ncumul, size=4)
    array([14.80803668, 44.95292261, 29.80797715, 23.11082091])
    """
    rng, _ = _rng_init(rng=rng, rng_seed=rng_seed, rng_state=rng_state, **kwargs)

    # Check that the shapes and sizes of diameters and cdf are compatible
    if np.isscalar(diameters) or np.isscalar(cdf):
        raise ValueError("The 'diameters' and 'cdf' arguments must be arrays")
    if diameters.ndim != 1:
        raise ValueError("The 'diameters' argument must be a 1-dimensional array")
    if diameters.size != cdf.shape[0]:
        raise ValueError("The 'diameters' array must have the same size as axis 0 of 'cdf'")
    if diameters.size < 2:
        raise ValueError("The 'diameters' and 'cdf' arguments must have at least two elements")
    if np.any(diameters <= 0.0):
        raise ValueError("All values in the 'diameters' argument must be positive")
    if np.any(cdf <= 0.0):
        raise ValueError("All values in the 'cdf' argument must be positive")
    if size is None and mu is None:
        size = 1
    elif size is not None and mu is not None:
        raise ValueError("The 'size' and 'mu' arguments are mutually exclusive")
    elif size is None and mu is not None:
        size = rng.poisson(mu)

    # Sort the diameters in descending order and get the cumulative distribution if it was not supplied
    sorted_indices = np.argsort(diameters)[::-1]
    sorted_diameters = diameters[sorted_indices]
    sorted_cdf = cdf[sorted_indices]

    # Check to make sure that the CDF is correctly specified so that as diameter is decreasing it is monotonically increasing
    is_monotonic_increasing = np.all(np.diff(sorted_cdf) >= 0)
    if not is_monotonic_increasing:
        raise ValueError("The CDF must be monotonically increasing with decreasing diameter.")

    # Normalize the cdf and put it in logspace
    sorted_cdf /= sorted_cdf[-1]
    log_sorted_cdf = np.log(sorted_cdf)

    # Generate uniform random numbers for the entire sample size
    u = rng.uniform(low=sorted_cdf[0], high=sorted_cdf[-1], size=size)
    u = np.log(u)

    # Handle the situation where u is a scalar (size is None and diameters is a scalar)
    if np.isscalar(u):
        u = np.array([u])

    # Flatten u to work with it as a 1D array
    original_shape = u.shape
    u = u.flatten()

    # Find the indices where the random numbers would be inserted to maintain order
    # Use the right side of the interval to avoid edge effects for when u == sorted_cdf[0]
    # Because rng.uniform returns values in the half-open interval [sorted_cdf[0], sorted_cdf[-1]), u will never be exactly equal to sorted_cdf[-1]
    indices = np.searchsorted(log_sorted_cdf, u, side="right")

    # Initialize the new_diameters array
    new_diameters = np.empty(u.shape)

    log_diam_low = np.log(sorted_diameters[indices - 1])
    log_diam_high = np.log(sorted_diameters[indices])
    log_nval_low = log_sorted_cdf[indices - 1]
    log_nval_high = log_sorted_cdf[indices]

    fractions = (u - log_nval_low) / (log_nval_high - log_nval_low)
    log_diam_interp = log_diam_low + fractions * (log_diam_high - log_diam_low)
    new_diameters = np.exp(log_diam_interp)

    # Reshape new_diameters to the original shape of u
    new_diameters = new_diameters.reshape(original_shape)

    # Add a small random noise to the diameters
    noise = 1e-8 * rng.uniform(size=new_diameters.shape)
    new_diameters *= 1 + noise
    return new_diameters


def get_random_velocity(
    vmean: np.float64,
    vescape: np.float64 | None = None,
    size: int | tuple[int, ...] = 1,
    rng: Generator | None = None,
    rng_seed: int | None = None,
    rng_state: dict | None = None,
    **kwargs: Any,
) -> NDArray[np.float64]:
    """
    Sample impact velocities from a Rayleigh distribution given a mean velocity. Optionally account for the escape velocity.

    Parameters
    ----------
    vmean : np.float64
        The mean velocity of the distribution.
    vescape : np.float64 | None, optional
        The escape velocity of the target body. If None, the escape velocity is not used in the calculation, and the disrtribution will be a maxwellian. If it is included, then the distribution will depend on the value of vmean. If vmean > vescape, then the distribution will adjusted so that the escape velocity is the minimum velocity by computing an encounter velocity then summing the encounter and escape velocities in quadrature. If vmean < vescape, then the distirbution will be a truncated maxwellian.
    size : int or tuple of ints, optional
        The number of samples to generate. If the shape is (m, n, k), then m * n * k samples are drawn. If size is None (the default), a single value is returned if `diameters` is a scalar, otherwise an array of samples is returned with the same size as `diameters`.
    rng : numpy.random.Generator | None
        |rng|
    rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
        |rng_seed|
    rng_state : dict, optional
        |rng_state|
    **kwargs : Any
        |kwargs|

    Returns
    -------
    ndarray
        An array of impact velocities (in m/s).
    """
    rng, _ = _rng_init(rng=rng, rng_seed=rng_seed, rng_state=rng_state, **kwargs)

    if vescape is not None:
        if vmean > vescape:
            vencounter = np.sqrt(vmean**2 - vescape**2)
            velocities = get_random_velocity(vencounter, rng=rng, size=size)
            return np.sqrt(velocities**2 + vescape**2)
        else:
            velocities = np.full(size, 2 * vescape)
            while True:
                nbad = np.sum(velocities > vescape)
                velocities[velocities > vescape] = get_random_velocity(vmean, rng=rng, size=nbad)
                if np.all(velocities < vescape):
                    return velocities

    sigma = vmean / np.sqrt(8 / np.pi)

    vx = rng.normal(0, sigma, size=size)
    vy = rng.normal(0, sigma, size=size)
    vz = rng.normal(0, sigma, size=size)
    velocities = np.sqrt(vx**2 + vy**2 + vz**2)

    return velocities


def bounded_norm(
    mean: FloatLike,
    scale: FloatLike,
    size: int | tuple[int, ...] = 1,
    rng: Generator | None = None,
    rng_seed: int | None = None,
    rng_state: dict | None = None,
    **kwargs: Any,
) -> FloatLike:
    """
    Sample from a truncated normal distribution that is bounded by 1-sigma stdev.

    Parameters
    ----------
    loc : FloatLike
       mean of the distribution
    scale : FloatLike
       standard deviation and bounds of the distribution
    size : int or tuple of ints, optional
        The number of samples to generate. If the shape is (m, n, k), then m * n * k samples are drawn. If size is None (the default), a single value is returned if `diameters` is a scalar, otherwise an array of samples is returned with the same size as `diameters`.
    rng : numpy.random.Generator | None
        |rng|
    rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
        |rng_seed|
    rng_state : dict, optional
        |rng_state|
    **kwargs : Any

    Returns
    -------
    float
       Truncated norm bounded by loc-scale, loc+scale
    """
    rng, _ = _rng_init(rng=rng, rng_seed=rng_seed, rng_state=rng_state, **kwargs)

    lower_bound = mean - scale
    upper_bound = mean + scale
    truncated_normal = truncnorm(
        (lower_bound - mean) / scale,
        (upper_bound - mean) / scale,
        loc=mean,
        scale=scale,
    )

    return truncated_normal.rvs(size, random_state=rng)



from __future__ import annotations

from abc import abstractmethod
from collections.abc import Callable
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal

import numpy as np
import pandas as pd
import xarray as xr
from cratermaker._cratermaker import counting_bindings
from geopandas import GeoSeries
from matplotlib.axes import Axes
from numpy.typing import ArrayLike
from shapely.ops import transform
from tqdm import tqdm
from vtk import vtkPolyData

from cratermaker import __version__ as cratermaker_version
from cratermaker.components.crater import Crater
from cratermaker.components.morphology import Morphology, MorphologyCrater
from cratermaker.constants import VECTOR_DRIVER_TO_EXTENSION_MAP, FloatLike
from cratermaker.core.base import ComponentBase, import_components

DRIVER_TO_EXTENSION_MAP = {"NETCDF": "nc", "SCC": "scc", "VTK": "vtp", "VTP": "vtp", "CSV": "csv", **VECTOR_DRIVER_TO_EXTENSION_MAP}


if TYPE_CHECKING:
    from cratermaker.components.surface import LocalSurface, Surface

_TALLY_LONG_NAME = "Unique crater identification number"

# The number of layers used for tagging faces with crater ids. This allows a single face to contain multiple crater ids
_N_LAYER = 8

# The minimum number of faces required in a region to perform crater counting, which corresponds to a roughly 6-8 pix diameter crater
_MIN_FACE_FOR_COUNTING = 100


class Counting(ComponentBase):
    """
    Base class for all crater counting models. It defines the interface for tallying the observable craters on a surface.
    """

    _registry: dict[str, Counting] = {}

    def __init__(
        self,
        surface: Surface | LocalSurface,
        Crater: type[Crater] | None = None,
        reset: bool = True,
        **kwargs: Any,
    ):
        """
        **Warning:** This object should not be instantiated directly. Instead, use the ``.maker()`` method.

        Parameters
        ----------
        surface : Surface | LocalSurface
            The surface or local surface view to be counted.
        Crater : type[Crater], optional
            The Crater class associated with this counting model. This is used to ensure that the correct variable properties for from a specialized Crater class are available (such as one associated with a Morphology class) when importing craters from file. If not supplied, then the base Crater class is used.
        reset : bool, optional
            Flag to indicate whether to reset the count and delete any old output files. Default is True.
        **kwargs : Any
            |kwargs|
        """
        from cratermaker.components.surface import Surface

        super().__init__(reset=reset, **kwargs)

        object.__setattr__(self, "_emplaced", [])
        object.__setattr__(self, "_observed", {})
        object.__setattr__(self, "_output_dir_name", "counting")
        object.__setattr__(self, "_output_file_prefix", "craters")
        object.__setattr__(self, "_output_file_extension", "nc")
        object.__setattr__(self, "_Crater", None)
        object.__setattr__(self, "_surface", None)
        object.__setattr__(self, "_morphology", None)
        object.__setattr__(
            self,
            "_driver_to_extension_map",
            {"SCC": "scc", "VTK": "vtk", "VTP": "vtp", "CSV": "csv", **VECTOR_DRIVER_TO_EXTENSION_MAP},
        )
        self._surface = Surface.maker(surface, reset=reset, **kwargs)
        self.Crater = Crater
        self._output_file_pattern += [
            f"observed_{self._output_file_prefix}*.{self._output_file_extension}",
            f"emplaced_{self._output_file_prefix}*.{self._output_file_extension}",
        ]

        if not reset:
            observed, emplaced = self.read_saved_output(interval=-1)
            if observed:
                interval = observed.interval.values[-1]
                observed = self.from_xarray(observed, interval=interval)
                for crater in observed:
                    self._observed[crater.id] = crater
                if emplaced:
                    self._emplaced = self.from_xarray(emplaced, interval=interval)
        return

    @classmethod
    def maker(
        cls,
        counting: str | Counting | None = None,
        surface: Surface | LocalSurface | None = None,
        Crater: type[Crater] | None = None,
        reset: bool = True,
        **kwargs: Any,
    ) -> Counting:
        """
        Initialize a Counting model with the given name or instance.

        Parameters
        ----------
        counting : str, Counting or None, default=None
            The name of the counting model to initialize. If None, the default model is used.
        surface : Surface | LocalSurface
            The surface or local surface view to be counted.
        Crater : type[Crater], optional
            The Crater class associated with this counting model. This is used to ensure that the correct variable properties for from a specialized Crater class are available (such as one associated with a Morphology class) when importing craters from file. If not supplied, then the base Crater class is used.
        reset : bool, optional
            Flag to indicate whether to reset the count and delete any old output files. Default is True
        **kwargs : Any
            |kwargs|

        Returns
        -------
        Counting
            An instance of the specified counting model.

        Raises
        ------
        KeyError
            If the specified counting model name is not found in the registry.
        TypeError
            If the specified counting model is not a string or a subclass of Scaling.
        """
        if counting is None:
            counting = "simplecount"

        counting = super().maker(
            component=counting,
            surface=surface,
            reset=reset,
            Crater=Crater,
            **kwargs,
        )

        return counting

    def __str__(self) -> str:
        base = super().__str__()
        str_repr = f"{base}\n"
        str_repr += f"Number of observed craters: {self.n_observed}\n"
        str_repr += f"Number of emplaced craters: {self.n_emplaced}\n"
        str_repr += f"\n{self.surface}\n"
        return str_repr

    def reset(self, **kwargs: Any) -> None:
        """
        Remove all craters count records from the surface and delete any output files.

        Parameters
        ----------
        **kwargs : Any
            |kwargs|
        """
        self.surface.add_tag(name="crater_id", long_name=_TALLY_LONG_NAME, tag=None, n_layer=self.n_layer)
        self._emplaced = []
        self._observed = {}

        super().reset(**kwargs)
        return

    def add(self, crater: MorphologyCrater, **kwargs: Any):
        """
        Add a crater to the surface.

        Parameters
        ----------
        crater : Crater
            The crater to be added to the surface.
        **kwargs : Any
            |kwargs|
        """
        if not isinstance(crater, MorphologyCrater):
            raise TypeError("crater must be an instance of MorphologyCrater")

        if self.surface.uxds is None:
            raise ValueError(
                "Surface must have an associated Uxarray dataset to use for counting. This is commonly caused by using a HiResLocal surface type without setting the superdomain_scale_factor."
            )

        # Tag a region just outside crater rim with the id
        crater_region = crater.crater_region

        if crater_region and crater_region.n_face >= _MIN_FACE_FOR_COUNTING:
            crater_region.add_tag(
                name="crater_id",
                long_name=_TALLY_LONG_NAME,
                tag=crater.id,
                **kwargs,
            )

            # Check to make sure the crater was recorded to the surface. HiResLocal surfaces may not record craters in the superdomain
            if crater.id in self.surface.uxds.crater_id:
                self.emplaced.append(crater)
                self.observed[crater.id] = crater
            else:
                return

            # Cookie cutting: remove any smaller craters that are overlapped by this new crater
            unique_ids = np.unique(crater_region.crater_id)
            unique_ids = unique_ids[unique_ids > 0]  # Remove the 0 id which corresponds to no crater
            if len(unique_ids) > 0:
                # Compute cookie cutting removes list
                observed = self.observed.copy()
                removes = [id for id, v in observed.items() if v.id in unique_ids and v.diameter < crater.diameter]
                # For every id that appears in the removes list, set it to 0 in the data array
                for remove_id in removes:
                    crater_region.remove_tag(name="crater_id", tag=remove_id)
                    if not np.any(
                        self.surface.uxds.crater_id.data == remove_id
                    ):  # Check to see if this crater id still appears, and if not, it's gone man.
                        self.observed.pop(remove_id, None)

        return

    def remove(self, crater_id: int) -> None:
        """
        Remove all instances of a crater id from the surface.

        Parameters
        ----------
        crater_id : int
            The ID of the crater to be removed.
        """
        self.surface.remove_tag(name="crater_id", tag=crater_id)
        self.observed.pop(crater_id, None)
        return

    def fit_rim(
        self, crater: Crater, tol=0.01, nloops=4, score_quantile=0.95, fit_center=False, fit_ellipse=False, **kwargs
    ) -> Crater:
        """
        Find the rim region of a crater on the surface.

        Parameters
        ----------
        crater : Crater
            The crater for which to find the rim region.
        tol : float, optional
            The tolerance for the rim fitting algorithm. Default is 0.01.
        nloops : int, optional
            The number of iterations for the rim fitting algorithm. Default is 4.
        score_quantile : float, optional
            The quantile of rim scores to consider. Default is 0.95.
        fit_center : bool, optional
            If True, fit the crater center as well. Default is False.
        fit_ellipse : bool, optional
            If True, fit an ellipse to the rim, otherwise fit a circle. Default is False.
        **kwargs : Any
            |kwargs|

        Returns
        -------
        Crater
            A new Crater object with updated measured rim parameters.
        """
        if not isinstance(crater, Crater):
            raise TypeError("crater must be an instance of Crater")

        location, ap, bp, orientation = counting_bindings.fit_rim(
            self.surface,
            crater,
            tol,
            nloops,
            score_quantile,
            fit_center,
            fit_ellipse,
        )

        crater.measured_semimajor_axis = ap
        crater.measured_semiminor_axis = bp
        crater.measured_orientation = np.degrees(orientation)
        crater.measured_location = location

        return crater

    def score_rim(self, crater: Crater, quantile=0.95, gradmult=1.0, curvmult=1.0, heightmult=1.0) -> None:
        """
        Score the rim region of a crater on the surface.

        Parameters
        ----------
        crater : Crater
            The crater for which to score the rim region.
        quantile : float, optional
            The quantile of rim scores to consider. Default is 0.95.
        gradmult : float, optional
            Gradient multiplier for scoring. Default is 1.0.
        curvmult : float, optional
            Curvature multiplier for scoring. Default is 1.0.
        heightmult : float, optional
            Height multiplier for scoring. Default is 1.0.

        Returns
        -------
        Updates the attached surface object with the rim score for this crater.
        """
        if not isinstance(crater, Crater):
            raise TypeError("crater must be an instance of Crater")

        region = crater.crater_region
        region.compute_desloped_face_elevation()
        region = counting_bindings.score_rim(region, crater, quantile, gradmult, curvmult, heightmult)
        return region

    def tally(self, region: LocalSurface | None = None, measure_rim: bool = False, quiet: bool = False, **kwargs: Any) -> None:
        """
        Tally the craters on the surface using the method of Minton et al. (2019) [#]_.

        Parameters
        ----------
        region : LocalSurface, optional
            A LocalSurface region to count. If not supplied, then the associated surface property is used.
        measure_rim : bool, optional
            If True, measure the rim of each crater. Default is False.
        quiet : bool, optional
            If True, suppress progress output. Default is False.
        **kwargs : Any
            |kwargs|

        References
        ----------
        .. [#] Minton, D.A., Fassett, C.I., Hirabayashi, M., Howl, B.A., Richardson, J.E., (2019). The equilibrium size-frequency distribution of small craters reveals the effects of distal ejecta on lunar landscape morphology. Icarus 326, 63-87. https://doi.org/10.1016/j.icarus.2019.02.021

        """
        if region is None:
            region = self.surface
            if hasattr(self.surface, "crater_id"):
                id_array = self.surface.crater_id
            else:
                return
        elif isinstance(region, LocalSurface):
            id_array = region.crater_id
        else:
            raise TypeError(f"Expected a LocalSurface, but got {type(region).__name__}.")

        unique_ids = np.unique(id_array[id_array > 0])
        remove_ids = []

        if quiet:
            iterable = unique_ids
        else:
            iterable = tqdm(
                unique_ids,
                total=len(unique_ids),
                desc="Counting craters",
                unit="craters",
                position=1,
                leave=False,
            )

        for id in iterable:
            # Check if we have orphaned crater ids for some reason and remove them
            if id not in self.observed:
                remove_ids.append(id)
                continue

            # Update the crater size measurement before computing the degradation and visibility functions
            crater = self.observed[id]
            # TODO: Make the fit_rim function more reliable before turning it on by default. It is currently very slow and not reliable enough.
            if measure_rim:
                fit_center = kwargs.pop("fit_center", False)
                fit_ellipse = kwargs.pop("fit_ellipse", False)
                crater = self.fit_rim(crater=crater, fit_center=fit_center, fit_ellipse=fit_ellipse, **kwargs)
            Kd = self.measure_degradation_state(crater, **kwargs)
            Kv = self.visibility_function(crater, **kwargs)
            if Kd >= Kv:
                remove_ids.append(id)
            else:
                self.observed[id] = crater  # Save the updated measurements to the observed tally

        if len(remove_ids) > 0:
            if quiet:
                iterable = remove_ids
            else:
                iterable = tqdm(
                    remove_ids,
                    total=len(remove_ids),
                    desc="Removing craters",
                    unit="craters",
                    position=2,
                    leave=False,
                )
            for id in iterable:
                self.remove(id)

        self.remove_complex_data()

        return

    @abstractmethod
    def measure_degradation_state(self, crater: Crater, **kwargs: Any) -> float: ...

    @abstractmethod
    def visibility_function(self, crater: Crater, Kv1: float = 0.17, gamma: float = 2.0, **kwargs: Any) -> float: ...

    @staticmethod
    def to_xarray(craters: dict[int, Crater] | list[Crater]) -> xr.Dataset:
        """
        Convert a list or dictionary of Crater objects to an xarray Dataset.

        Parameters
        ----------
        craters : dict[int, Crater] | list[Crater]
            A dictionary or list of Crater objects to convert.

        Returns
        -------
        xr.Dataset
            An xarray Dataset containing the crater data.
        """
        if len(craters) == 0:
            return xr.Dataset()
        new_data = []
        if isinstance(craters, dict):
            craters = craters.values()
        for c in craters:
            d = c.as_dict(skip_complex_data=True)
            if "location" in d:
                lon, lat = d.pop("location")
                d["longitude"] = lon
                d["latitude"] = lat
            if "measured_location" in d:
                mlon, mlat = d.pop("measured_location")
                d["measured_longitude"] = mlon
                d["measured_latitude"] = mlat
            d = xr.Dataset(data_vars=d).set_coords("id").expand_dims(dim="id")
            d["id"].attrs["long_name"] = _TALLY_LONG_NAME
            new_data.append(d)

        return xr.concat(new_data, dim="id")

    def _to_file(
        self,
        craters: dict[int, Crater] | list[Crater],
        filename: Path | str,
        interval: int = 0,
    ) -> xr.Dataset | None:
        """
        Merge a list or dictionary of Crater objects with an existing file.

        Parameters
        ----------
        craters : dict[int, Crater] | list[Crater]
            A dictionary or list of Crater objects to merge.
        filename : Path | str
            The path to the file to merge with.
        interval : int, optional
            The interval number. This is added to the coordinates of the dataset created from the craters before merging with the file. Default is 0.

        Returns
        -------
        xr.Dataset | None
            An xarray Dataset containing the merged crater data, or None if no data.
        """
        # Convert into an xarray dataset
        combined_data = self.to_xarray(craters)
        combined_data = combined_data.expand_dims(dim="interval").assign_coords({"interval": [interval]})

        # If the file already exists, read it and merge
        if filename.exists():
            filename.unlink()

        # Write merged data back to file
        if combined_data:
            combined_data.to_netcdf(filename)
        return combined_data

    def save(
        self,
        crater_type: Literal["observed", "emplaced", "both"] = "both",
        interval: int = 0,
        craters: list[Crater] | None = None,
        **kwargs: Any,
    ) -> None:
        """
        Dump the crater lists to a file and reset the emplaced crater list.

        Parameters
        ----------
        crater_type : str, optional
            The type of craters to export. Options are "observed", "emplaced", and "both". Default is "observed".
        interval : int, default=0
            The interval number for the output file naming.
        craters: list[Crater] | None, optional
            An arbitrary list of craters to save. If None, then the current observed and/or emplaced tallies are used. Default is None.
        **kwargs : Any
            |kwargs|
        """
        if craters is None:
            self.remove_complex_data()
            observed = self.observed
            emplaced = self.emplaced
        else:
            if crater_type == "both":
                raise ValueError(
                    "If supplying a custom list of craters to save, then crater_type must be either 'observed' or 'emplaced', not 'both'."
                )
            elif crater_type == "observed":
                observed = {crater.id: crater for crater in craters}
                emplaced = None
            elif crater_type == "emplaced":
                observed = None
                emplaced = craters

        if crater_type == "observed":
            iter = zip([observed], ["observed"], strict=True)
        elif crater_type == "emplaced":
            iter = zip([emplaced], ["emplaced"], strict=True)
        elif crater_type == "both":
            iter = zip([observed, emplaced], ["observed", "emplaced"], strict=True)
        else:
            raise ValueError(f"Invalid crater_type: {crater_type}. Must be one of 'observed', 'emplaced', or 'both'.")
        for craters, name in iter:
            if craters:
                filename = self.output_dir / f"{name}_{self.output_filename(interval)}"
                self._to_file(craters, filename, interval)
        save_args = {"interval": interval, **kwargs}
        super().save(**save_args)
        return

    def from_file(self, filename: str | Path, **kwargs: Any) -> None:
        pass

    def export(
        self,
        crater_type: Literal["observed", "emplaced", "both"] = "observed",
        interval: int | None = None,
        driver: str = "SCC",
        ask_overwrite: bool | None = None,
        **kwargs: Any,
    ) -> None:
        """
        Exports crater lists to a file.

        Parameters
        ----------
        crater_type : str, optional
            The type of craters to export. Options are "observed", "emplaced", and "both". Default is "observed".
        interval : int | None, optional
            |interval_export|
        driver : str, default='SCC'
            The export driver used to save. The valid drivers are given by the `valid_drivers` attribute of this object.
        ask_overwrite : bool, optional
            |ask_overwrite_methods|
        **kwargs : Any
            |kwargs|
        """
        # Check if the requested driver is a supported one for crater counts, and return silently if not.
        if driver.upper() not in self.valid_drivers:
            return

        # Temporarily set the ask_overwrite attribute for the duration of the export, but reset it to its original value afterwards.
        ask_overwrite_orig = self.ask_overwrite
        if ask_overwrite is not None:
            self.ask_overwrite = ask_overwrite

        crater_names = ["observed", "emplaced"]
        output_ds = self.read_saved_output(interval=interval)
        for name, crater_ds in zip(crater_names, output_ds, strict=True):
            if crater_type == "observed" and name != "observed":
                continue
            if crater_type == "emplaced" and name != "emplaced":
                continue

            if type(crater_ds) is dict:
                interval_numbers = list(crater_ds.keys())
            elif crater_ds is None or "interval" not in crater_ds:
                if interval is None:
                    interval_numbers = [0]
                else:
                    interval_numbers = [interval]
            else:
                interval_numbers = crater_ds.interval.values
            for interval in interval_numbers:
                if driver.upper() == "NETCDF" and crater_ds is not None:
                    self.save(crater_type=name, interval=interval, craters=self.from_xarray(crater_ds, interval=interval), **kwargs)
                elif driver.upper() in ["VTK", "VTP"] and crater_ds is not None:
                    self.to_vtk_file(
                        crater_ds=crater_ds,
                        interval=interval,
                        name=name,
                        **kwargs,
                    )
                elif driver.upper() == "CSV" and crater_ds is not None:
                    self.to_csv_file(
                        crater_ds=crater_ds,
                        interval=interval,
                        name=name,
                        **kwargs,
                    )
                elif driver.upper() == "SCC":
                    if crater_ds is None:
                        crater_ds = []  # Allow for empty SCC files because they can still have the region polygon
                    self.to_scc_file(
                        crater_ds=crater_ds,
                        interval=interval,
                        name=name,
                        **kwargs,
                    )
                elif driver.upper() in VECTOR_DRIVER_TO_EXTENSION_MAP and crater_ds is not None:
                    self.to_vector_file(
                        crater_ds=crater_ds,
                        interval=interval,
                        name=name,
                        driver=driver,
                        **kwargs,
                    )

        self.ask_overwrite = ask_overwrite_orig
        return

    def plot(
        self,
        interval: int | None = None,
        observed_color: str | None = "white",
        observed_original_color: str | None = None,
        emplaced_color: str | None = None,
        plot_style: Literal["map", "hillshade"] = "map",
        variable_name: str | None = None,
        cmap: str | None = None,
        show=True,
        save=True,
        ax: Axes | None = None,
        **kwargs: Any,
    ) -> Axes:
        """
        Plot an image of the local region.

        Parameters
        ----------
        interval : int | None, optional
            The interval number to load the emplaced crater data from. if None, then all emplaced data currently saved to file is used. Default is None.
        observed_color : str | None, optional
            The color to use for observed craters using their measured properties. If None, observed craters will not be plotted. Default is "white".
        observed_original_color : str | None, optional
            The color to use for observed craters using their original properties. If None, observed craters will not be plotted. Default is None.
        emplaced_color : str | None, optional
            The color to use for emplaced craters. If None, emplaced craters will not be plotted. Default is None.
        plot_style : str, optional
            The style of the plot. Options are "map" and "hillshade". In "map" mode, the variable is displayed as a colored map. In "hillshade" mode, a hillshade image is generated using "face_elevation" data. If a different variable is passed to `variable`, then the hillshade will be overlayed with that variable's data. Default is "map".
        variable_name : str | None, optional
            The variable to plot. If None is provided then "face_elevation" is used in "map" mode.
        cmap : str, optional
            The colormap to use for the plot. If None, a default colormap will be used ("cividis" by default and "grey" when plot_style=="hillshade" and variable=="face_elevation").
        show : bool, optional
            If True, the plot will be displayed. Default is True.
        save : bool, optional
            If True, the plot will be saved to the default plot directory. Default is True
        ax : matplotlib.axes.Axes, optional
            An existing Axes object to plot on. If None, a new figure and axes will be created.
        **kwargs : Any
            |kwargs|

        Returns
        -------
        matplotlib.image.Axes object created by the surface plot method, with crater counts plotted on top if specified. If show is True, the plot will also be displayed.
            The Axes object
        """
        import matplotlib.pyplot as plt

        from cratermaker.components.surface.hireslocal import HiResLocalSurface

        crs = self.surface.crs
        split_antimeridian = True
        file_prefix = f"{self.surface.output_file_prefix}"
        # Handle the HiResLocal surface case where we may or may not be plotting the global surface
        if isinstance(self.surface, HiResLocalSurface):
            superdomain = kwargs.pop("superdomain", False)
            if not superdomain and self.surface.local is not None:
                crs = self.surface.local.crs
                split_antimeridian = False
                file_prefix = f"{self.surface.local.output_file_prefix}"

        file_prefix += f"_{self.output_file_prefix}"
        if interval is not None:
            observed, emplaced = self.read_saved_output(interval=interval)
            if observed:
                interval = observed.interval.values[-1]
                observed = self.from_xarray(observed, interval=interval)
            if emplaced:
                emplaced_interval = emplaced.interval.values[-1]
                if emplaced_interval == interval:
                    emplaced = self.from_xarray(emplaced, interval=interval)
            filename = self.plot_dir / f"{file_prefix}{interval:06d}.{self.surface.output_image_file_extension}"
        else:
            observed = [c for _, c in self.observed.items()]
            emplaced = self.emplaced
            filename = self.plot_dir / f"{file_prefix}.{self.surface.output_image_file_extension}"
        if ax is None:
            W, H = self.surface.get_raster_dims()
            _, ax = plt.subplots(figsize=(1, 1), dpi=W, frameon=False)
        ax = self.surface.plot(
            interval=interval,
            show=False,
            save=False,
            ax=ax,
            plot_style=plot_style,
            variable_name=variable_name,
            cmap=cmap,
            **kwargs,
        )

        if emplaced_color is not None and emplaced is not None and len(emplaced) > 0:
            gs = self.to_geoseries(
                craters=emplaced, use_measured_properties=False, split_antimeridian=split_antimeridian, autolim=False
            ).to_crs(crs)
            facecolor = kwargs.pop("facecolor", "none")
            edgecolor = emplaced_color
            linewidth = kwargs.pop("linewidth", 0.1)
            linestyle = kwargs.pop("linestyle", "solid")
            ax = gs.plot(ax=ax, facecolor=facecolor, edgecolor=edgecolor, linewidth=linewidth, linestyle=linestyle)

        if observed_original_color is not None and observed is not None and len(observed) > 0:
            gs = self.to_geoseries(
                craters=observed, use_measured_properties=False, split_antimeridian=split_antimeridian, autolim=False
            ).to_crs(crs)
            facecolor = kwargs.pop("facecolor", "none")
            edgecolor = observed_original_color
            linewidth = kwargs.pop("linewidth", 0.1)
            linestyle = kwargs.pop("linestyle", "solid")
            ax = gs.plot(ax=ax, facecolor=facecolor, edgecolor=edgecolor, linewidth=linewidth, linestyle=linestyle)

        if observed_color is not None and observed is not None and len(observed) > 0:
            gs = self.to_geoseries(
                craters=observed, use_measured_properties=True, split_antimeridian=split_antimeridian, autolim=False
            ).to_crs(crs)
            facecolor = kwargs.pop("facecolor", "none")
            edgecolor = observed_color
            linewidth = kwargs.pop("linewidth", 0.1)
            linestyle = kwargs.pop("linestyle", "solid")
            ax = gs.plot(ax=ax, facecolor=facecolor, edgecolor=edgecolor, linewidth=linewidth, linestyle=linestyle)

        if save:
            print(f"Saving crater count plot to {filename}...")
            plt.savefig(filename, bbox_inches="tight", pad_inches=0, dpi=W)
        if show:
            plt.show()
        else:
            plt.close()
        return ax

    def show_pyvista(
        self,
        surface: Surface | LocalSurface | None = None,
        observed_color: str = "white",
        emplaced_color: str = "red",
        interval: int | None = None,
        **kwargs: Any,
    ):
        """
        Passes through to the surface show_pyvista method and adds crater counts to it.

        Parameters
        ----------
        surface : Surface | LocalSurface, optional
            The surface or local surface view to be displayed. If None, uses the associated surface property
        observed_color : str, optional
            The color to use for observed craters. Default is "white".
        emplaced_color : str, optional
            The color to use for emplaced craters. Default is "red".
        interval : int, optional
            The interval number to load the emplaced crater data from. if None, then all emplaced data currently saved to file is used. Default is None.
        **kwargs : Any
            |kwargs|

        Returns
        -------
        plotter : pyvista.Plotter
            The pyvista plotter with the crater counts added.

        """
        if surface is None:
            surface = self.surface
        plotter = surface.show_pyvista(**kwargs)

        from cratermaker.utils.general_utils import toggle_pyvista_actor, update_pyvista_help_message

        if interval is None:
            interval = -1

        observed, emplaced = self.read_saved_output(interval=interval)
        if observed:
            interval = observed.interval.values[-1]
            observed = self.from_xarray(observed, interval=interval)
            observed_count_actor = plotter.add_mesh(
                self.to_vtk_mesh(observed, use_measured_properties=True), line_width=2, color=observed_color, name="observed"
            )
            observed_count_actor.SetVisibility(False)
            plotter.add_key_event("c", lambda: toggle_pyvista_actor(plotter, observed_count_actor))
            plotter = update_pyvista_help_message(plotter, new_message="c: Toggle counted craters")
        if emplaced:
            emplaced_interval = emplaced.interval.values[-1]
            if emplaced_interval == interval:
                emplaced = self.from_xarray(emplaced, interval=interval)
                emplaced_count_actor = plotter.add_mesh(
                    self.to_vtk_mesh(emplaced, use_measured_properties=False),
                    line_width=2,
                    color=emplaced_color,
                    name="emplaced",
                )
                emplaced_count_actor.SetVisibility(False)
                plotter.add_key_event("t", lambda: toggle_pyvista_actor(plotter, emplaced_count_actor))
                plotter = update_pyvista_help_message(plotter, new_message="t: Toggle emplaced craters")

        return plotter

    def show3d(self, engine: str = "pyvista", observed_color: str = "white", emplaced_color: str = "red", **kwargs: Any) -> None:
        """
        Passes through to the surface show method and adds crater counts to it.

        Parameters
        ----------
        engine : str, optional
            The engine to use for plotting. Currently, only "pyvista" is supported. Default is "pyvista".
        observed_color : str, optional
            The color to use for observed craters. Default is "white".
        emplaced_color : str, optional
            The color to use for emplaced craters. Default is "red".
        **kwargs : Any
            |kwargs|
        """
        if engine.lower() == "pyvista":
            plotter = self.show_pyvista(observed_color=observed_color, emplaced_color=emplaced_color, **kwargs)
            plotter.show()
        else:
            raise ValueError(f"Engine '{engine}' is not supported for crater counting visualization.")
        return

    def to_geoseries(
        self,
        craters: list[Crater] | None = None,
        use_measured_properties: bool = True,
        split_antimeridian: bool = False,
        **kwargs: Any,
    ) -> GeoSeries:
        if craters is None:
            craters = [c for _, c in self.observed.items()]
        surface = self.surface
        crater_gs = []
        for crater in tqdm(
            craters,
            total=len(craters),
            desc="Converting craters to GeoSeries polygons",
            unit="crater",
            position=0,
            leave=False,
        ):
            crater_gs.append(
                crater.to_geoseries(surface=surface, split_antimeridian=split_antimeridian, use_measured_properties=False)
            )
        return pd.concat(crater_gs)

    def _validate_export_args(
        self,
        name: Literal["observed", "emplaced"] = "observed",
        interval: int | None = None,
        crater_ds: xr.Dataset | list[Crater] | dict[int, Crater] | None = None,
    ) -> list[Crater]:
        if crater_ds is None:
            crater_list = getattr(self, name)
        if type(crater_ds) is dict:
            if type(list(crater_ds.values())[0]) is xr.Dataset:  # like multi-interval dicts keyed by interval
                crater_ds = crater_ds.get(
                    interval, []
                )  # Get the dataset for the specified interval, or an empty dataset if not found
            else:
                return list(crater_ds.values())  # like observed dicts keyed by crater id

        if type(crater_ds) is list:
            crater_list = crater_ds
        elif type(crater_ds) is dict:
            crater_list = list(crater_ds.values())
        elif type(crater_ds) is xr.Dataset:
            crater_list = self.from_xarray(crater_ds, interval=interval)
        else:
            raise ValueError(f"Unrecognized type for crater_ds: {type(crater_ds)}")

        return crater_list

    def to_vector_file(
        self,
        name: Literal["observed", "emplaced"] = "observed",
        crater_ds: xr.Dataset | list[Crater] | dict[int, Crater] | None = None,
        interval: int | None = None,
        driver: str = "GPKG",
        use_measured_properties: bool = True,
        **kwargs,
    ) -> None:
        """
        Export the crater data to a vector file and stores it in the default export directory.

        Parameters
        ----------
        name : Literal["observed", "emplaced"], optional
            The name of the crater dataset to export, either "observed" or "emplaced
        crater_ds : xr.Dataset | list[Crater] | dict[int, Crater] | None, optional
            The crater data to export. Can be provided as an xarray Dataset, a list of Crater objects, or a dictionary mapping interval numbers to Crater objects. If None, the crater data will be the attribute of the class corresponding to the name parameter (self.observed or self.emplaced). Default is None.
        interval : int | None, optional
            |interval_export|
        driver : str, optional
            The file format to save. Supported formats are 'GPKG', 'ESRI Shapefile', etc. Default is 'GPKG'.
        use_measured_properties : bool, optional
            If True, use the current measured crater properties (semimajor_axis, semiminor_axis, location, orientation) instead of the initial ones, by default True.
        **kwargs : Any
            |kwargs|
        """
        return
        # TODO Fix this
        # from cratermaker.constants import VECTOR_DRIVER_TO_EXTENSION_MAP

        # crater_list = self._validate_export_args(name=name, interval=interval, crater_ds=crater_ds)

        # def shp_key_fix(key: str) -> str:
        #     """
        #     ESRI Shapefile format limits field names to 10 characters, so this function substitues longer names with shorter alternatives, truncates the results, and sets them to upper case.
        #     """
        #     alt_names = {
        #         "projectile_": "proj",
        #         "morphology_": "morph",
        #         "diameter": "diam",
        #         "longitude": "lon",
        #         "latitude": "lat",
        #         "density": "dens",
        #         "velocity": "vel",
        #         "direction": "dir",
        #         "location": "loc",
        #         "angle": "ang",
        #         "transient_": "tr",
        #         "semimajor_axis": "a",
        #         "semiminor_axis": "b",
        #         "orientation": "orient",
        #         "measured_": "meas",
        #         "degradation_state": "kdeg",
        #     }
        #     for long, short in alt_names.items():
        #         if long in key:
        #             key = key.replace(long, short)
        #     return key[:10].upper()

        # # Common alias for Shapefile
        # if driver.upper() == "SHP":
        #     driver = "ESRI Shapefile"
        # if driver.upper() in VECTOR_DRIVER_TO_EXTENSION_MAP:
        #     file_extension = VECTOR_DRIVER_TO_EXTENSION_MAP[driver.upper()]
        # else:
        #     raise ValueError("Cannot infer file extension from driver {driver}.")

        # if file_extension == "shp":
        #     format_has_layers = False
        # else:
        #     format_has_layers = True

        # surface = self.surface
        # split_antimeridian = False

        # geoms = []
        # attrs = []
        # for crater in tqdm(
        #     crater_list,
        #     total=len(crater_list),
        #     desc=f"Converting {name} craters to geometries for export",
        #     unit="craters",
        #     position=0,
        #     leave=False,
        # ):
        #     poly = crater.to_geoseries(
        #         surface=surface, split_antimeridian=split_antimeridian, use_measured_properties=use_measured_properties
        #     ).item()
        #     df = crater_ds.sel(id=[crater.id]).to_dataframe()
        #     if isinstance(poly, GeometryCollection):
        #         for p in poly.geoms:
        #             geoms.append(p)
        #             attrs.append(df)
        #     else:
        #         geoms.append(poly)
        #         attrs.append(df)

        # if len(geoms) > 0:
        #     attrs_df = pd.concat(attrs, ignore_index=True)
        # else:
        #     attrs_df = pd.DataFrame()
        # if driver.upper() == "ESRI SHAPEFILE":
        #     attrs_df.rename(mapper=shp_key_fix, axis=1, inplace=True)

        # gdf = gpd.GeoDataFrame(data=attrs_df, geometry=geoms, crs=surface.crs)
        # if format_has_layers:
        #     output_file = self.export_dir / f"{self.output_file_prefix}{interval:06d}.{file_extension}"
        #     print(f"Saving {name} layer to vector file: '{output_file}'...")
        # else:
        #     output_file = self.export_dir / f"{name}_{self.output_file_prefix}{interval:06d}.{file_extension}"
        #     if not self._overwrite_check(output_file):
        #         return
        # if driver.upper() == "ESRI SHAPEFILE" and hasattr(self.surface, "local"):
        #     # Create the _AREA file
        #     self.surface.local.export_region_polygon(driver=driver)

        # try:
        #     if format_has_layers:
        #         gdf.to_file(output_file, layer=name)
        #     else:
        #         gdf.to_file(output_file)
        # except Exception as e:
        #     raise RuntimeError(f"Error saving {output_file}: {e}") from e

        # return

    def to_vtk_mesh(self, craters: list[Crater], use_measured_properties: bool = True, **kwargs: Any) -> vtkPolyData:
        """
        Convert the crater data to a VTK PolyData mesh.

        Parameters
        ----------
        craters : list[Crater]
            A list of Crater objects to convert.
        use_measured_properties : bool, optional
            If True, use the current measured crater properties (semimajor_axis, semiminor_axis, location, orientation) instead of the initial ones, by default True.
        **kwargs : Any
            |kwargs|

        Returns
        -------
        vtkPolyData
            A VTK PolyData object representing the crater geometries.
        """
        from vtk import (
            vtkCellArray,
            vtkPoints,
            vtkPolyLine,
            vtkXMLPolyDataWriter,
        )

        def lonlat_to_xyz(R):
            def _f(lon_deg, lat_deg, z=0.0):
                lon = np.deg2rad(lon_deg)
                lat = np.deg2rad(lat_deg)
                X = (R + z) * np.cos(lat) * np.cos(lon)
                Y = (R + z) * np.cos(lat) * np.sin(lon)
                Z = (R + z) * np.sin(lat)
                return X, Y, Z

            return _f

        def polygon_xyz_coords(geom, R):
            """Yield Nx3 arrays for each exterior ring (drop closing vertex)."""
            g3d = transform(lonlat_to_xyz(R), geom)

            def _rings(g):
                if g.geom_type == "Polygon":
                    yield np.asarray(g.exterior.coords)[:-1]  # (N, 3)
                    for i in g.interiors:
                        yield np.asarray(i.coords)[:-1]
                elif g.geom_type in ("MultiPolygon", "GeometryCollection"):
                    for sub in g.geoms:
                        yield from _rings(sub)
                else:
                    yield np.asarray(g.coords)

            yield from _rings(g3d)

        surface = self.surface

        geoms = []
        for crater in craters:
            geoms.append(
                crater.to_geoseries(
                    surface=surface, split_antimeridian=False, use_measured_properties=use_measured_properties, **kwargs
                )
            )

        points = vtkPoints()
        lines = vtkCellArray()
        point_id = 0  # Keep track of the point ID across all circles
        for g in geoms:
            for ring_xyz in polygon_xyz_coords(g.item(), surface.radius):
                x, y, z = ring_xyz.T  # each is (N,)
                for i in range(len(x)):
                    points.InsertNextPoint(float(x[i]), float(y[i]), float(z[i]))
                polyline = vtkPolyLine()
                polyline.GetPointIds().SetNumberOfIds(len(x))
                for i in range(len(x)):
                    polyline.GetPointIds().SetId(i, point_id + i)
                point_id += len(x)
                lines.InsertNextCell(polyline)

        # Create a poly_data object and add points and lines to it
        poly_data = vtkPolyData()
        poly_data.SetPoints(points)
        poly_data.SetLines(lines)
        return poly_data

    def to_vtk_file(
        self,
        name: Literal["observed", "emplaced"] = "observed",
        crater_ds: xr.Dataset | list[Crater] | dict[int, Crater] | None = None,
        interval: int | None = None,
        **kwargs,
    ) -> None:
        """
        Export the crater data to a VTK file and stores it in the default export directory.

        Notes: In order for the crater and surface to be synced up when saving to VTK/VTP format, the initial conditions (no craters) must be saved. Otherwise, saving to file only occurs if there are craters to save.

        Parameters
        ----------
        name : Literal["observed", "emplaced"], optional
            The name of the crater dataset to export, either "observed" or "emplaced
        crater_ds : xr.Dataset | list[Crater] | dict[int, Crater] | None, optional
            The crater data to export. Can be provided as an xarray Dataset, a list of Crater objects, or a dictionary mapping interval numbers to Crater objects. If None, the crater data will be the attribute of the class corresponding to the name parameter (self.observed or self.emplaced). Default is None.
        interval : int | None, optional
           |interval_export|
        **kwargs : Any
            |kwargs|
        """
        from vtk import vtkXMLPolyDataWriter

        crater_list = self._validate_export_args(name=name, interval=interval, crater_ds=crater_ds)

        filename_base = self.output_filename(interval).replace(self.output_file_extension, "vtp")
        output_file = self.export_dir / f"{name}_{filename_base}"
        if not self._overwrite_check(output_file):
            return
        print(f"Saving crater data to VTK file: '{output_file}'...")
        poly_data = self.to_vtk_mesh(craters=crater_list)

        # Write the poly_data to a VTK file
        writer = vtkXMLPolyDataWriter()
        writer.SetFileName(output_file)
        writer.SetInputData(poly_data)

        # Optional: set the data mode to binary to save disk space
        writer.SetDataModeToBinary()
        writer.Write()
        return

    def to_csv_file(
        self,
        name: Literal["observed", "emplaced"] = "observed",
        crater_ds: xr.Dataset | list[Crater] | dict[int, Crater] | None = None,
        interval: int | None = None,
        **kwargs,
    ) -> None:
        """
        Export the crater data to a CSV file and stores it in the default export directory.

        Parameters
        ----------
        name : Literal["observed", "emplaced"], optional
            The name of the crater dataset to export, either "observed" or "emplaced
        crater_ds : xr.Dataset | list[Crater] | dict[int, Crater] | None, optional
            The crater data to export. Can be provided as an xarray Dataset, a list of Crater objects, or a dictionary mapping interval numbers to Crater objects. If None, the crater data will be the attribute of the class corresponding to the name parameter (self.observed or self.emplaced). Default is None.
        interval : int | None, optional
            |interval_export|
        **kwargs : Any
            |kwargs|
        """
        import csv

        crater_list = self._validate_export_args(name=name, interval=interval, crater_ds=crater_ds)

        filename_base = self.output_filename(interval).replace(self.output_file_extension, "csv")
        output_file = self.export_dir / f"{name}_{filename_base}"
        if not self._overwrite_check(output_file):
            return
        print(f"Saving crater data to CSV file: '{output_file}'...")
        with output_file.open(mode="w", newline="") as csvfile:
            writer = csv.writer(csvfile)
            header_written = False
            for crater in crater_list:
                crater_dict = crater.as_dict(skip_complex_data=True)

                # Convert location fields from tuples into lon/lat
                location = crater_dict.pop("location")
                crater_dict["longitude"] = location[0]
                crater_dict["latitude"] = location[1]
                measured_location = crater_dict.pop("measured_location")
                crater_dict["measured_longitude"] = measured_location[0]
                crater_dict["measured_latitude"] = measured_location[1]

                # Check if the crater is circular, and if so convert semimajor/semiminor to diameter
                if (
                    "measured_semimajor_axis" in crater_dict
                    and "measured_semiminor_axis" in crater_dict
                    and crater_dict["measured_semimajor_axis"] == crater_dict["measured_semiminor_axis"]
                ):
                    measured_diameter = 2.0 * crater_dict.pop("measured_semimajor_axis")
                    crater_dict.pop("measured_semiminor_axis")
                    items = list(crater_dict.items())
                    crater_dict = {"id": crater_dict["id"], "measured_diameter": measured_diameter}
                    crater_dict.update(items[1:])

                if crater_dict["semimajor_axis"] == crater_dict["semiminor_axis"]:
                    diameter = 2.0 * crater_dict.pop("semimajor_axis")
                    crater_dict.pop("semiminor_axis")
                    items = list(crater_dict.items())
                    crater_dict = {"id": crater_dict["id"], "diameter": diameter}
                    crater_dict.update(items[1:])

                # Pop out any None values
                crater_dict = {k: v for k, v in crater_dict.items() if v is not None}

                if not header_written:
                    header = list(crater_dict.keys())
                    writer.writerow(header)
                    header_written = True
                row = [crater_dict[key] for key in header]
                writer.writerow(row)

        return

    def from_csv_file(self, input_file: Path | str) -> list[Crater]:
        """
        Import crater data from a CSV file.

        Parameters
        ----------
        input_file : Path | str
            The path to the CSV file containing crater data.

        Returns
        -------
        list[Crater]
            A list of Crater objects imported from the CSV file.
        """
        import csv

        craters = []
        input_file = Path(input_file)
        if not input_file.exists():
            raise FileNotFoundError(f"Input file '{input_file}' does not exist.")
        with input_file.open(mode="r", newline="") as csvfile:
            reader = csv.DictReader(csvfile)
            for row in reader:
                crater_data = {}
                crater_data["location"] = [None, None]
                crater_data["measured_location"] = [None, None]
                for key, value in row.items():
                    if key in ["id"]:
                        crater_data[key] = int(value)
                    elif key == "longitude":
                        crater_data["location"][0] = float(value)
                    elif key == "latitude":
                        crater_data["location"][1] = float(value)
                    elif key == "measured_longitude":
                        crater_data["measured_location"][0] = float(value)
                    elif key == "measured_latitude":
                        crater_data["measured_location"][1] = float(value)
                    else:
                        try:
                            crater_data[key] = float(value)
                        except ValueError:
                            crater_data[key] = value
                if None in crater_data["location"]:
                    crater_data.pop("location")
                if None in crater_data["measured_location"]:
                    crater_data.pop("measured_location")
                crater = self.Crater.maker(**crater_data, morphology=self.morphology, check_redundant_inputs=False)
                craters.append(crater)

        return craters

    def to_scc_file(
        self,
        name: Literal["observed", "emplaced"] = "observed",
        crater_ds: xr.Dataset | list[Crater] | dict[int, Crater] | None = None,
        interval: int | None = None,
        **kwargs,
    ) -> None:
        """
        Export the crater data to Craterstats and OpenCraterTools-compatible SCC file and stores it in the default export directory.

        Parameters
        ----------
        name : Literal["observed", "emplaced"], optional
            The name of the crater dataset to export, either "observed" or "emplaced
        crater_ds : xr.Dataset | list[Crater] | dict[int, Crater] | None, optional
            The crater data to export. Can be provided as an xarray Dataset, a list of Crater objects, or a dictionary mapping interval numbers to Crater objects. If None, the crater data will be the attribute of the class corresponding to the name parameter (self.observed or self.emplaced). Default is None.
        interval : int | None, optional
            |interval_export|
        **kwargs : Any
            |kwargs|
        """
        import datetime

        crater_list = self._validate_export_args(name=name, interval=interval, crater_ds=crater_ds)

        def overlap_fraction(crater, region_poly=None):
            if region_poly is None:
                return 1.0
            distance = self.surface.compute_distances(reference_location=self.surface.local_location, locations=[crater.location])
            if distance + crater.measured_radius > self.surface.local_radius:
                crater_poly = crater.to_geoseries(
                    surface=self.surface, split_antimeridian=False, use_measured_properties=True
                ).to_crs(self.surface.crs)
                overlap_area = crater_poly.intersection(region_poly).to_crs(self.surface.local.crs).area.item()
                return overlap_area / crater_poly.to_crs(self.surface.local.crs).area.item()
            else:
                return 1.0

        region_poly = None

        output_file = self.export_dir / f"{name}{interval:06d}.scc"
        if not self._overwrite_check(output_file):
            return
        print(f"Saving crater data to {output_file}")
        with output_file.open(mode="w") as f:
            f.write(f"# Spatial crater count Cratermaker version {cratermaker_version}\n")
            f.write("#\n")
            f.write(f"# Exported on {datetime.datetime.now().isoformat()}\n")
            f.write("#\n")
            f.write("# Ellipsoid axes\n")
            f.write(f"a-axis radius = {self.surface.radius * 1e-3:.3f} <km>\n")
            f.write(f"b-axis radius = {self.surface.radius * 1e-3:.3f} <km>\n")
            f.write(f"c-axis radius = {self.surface.radius * 1e-3:.3f} <km>\n")

            # Start with regional area
            boundary_points = []

            if hasattr(self.surface, "local_radius") and hasattr(self.surface, "local_location"):
                f.write(f"coordinate_system_name = {self.surface.local.crs.name}\n")
                if region_poly is None:  # We only need to do this the first time through
                    region_circle = Crater.maker(
                        radius=self.surface.local_radius, location=self.surface.local_location
                    )  # We can get away with using just the base class for Crater here
                    region_poly = (
                        region_circle.to_geoseries(surface=self.surface, split_antimeridian=False, use_measured_properties=False)
                        .to_crs(self.surface.crs)
                        .item()
                    )
                    boundary_points = list(region_poly.exterior.coords)
                    area = self.surface.local.area
            else:
                f.write(f"coordinate_system_name = {self.surface.crs.name}\n")
                boundary_points = [(-180.0, -90.0), (180.0, -90.0), (180.0, 90.0), (-180.0, 90.0), (-180.0, -90.0)]
                area = self.surface.area
                region_poly = None
            f.write("# area_shapes:\n")
            f.write("unit_boundary = {vertex, sub_area, tag, lon, lat\n")
            for i, p in enumerate(boundary_points):
                f.write(f"{i}\t1\text\t{p[0]}\t{p[1]}\n")
            f.write("}\n")
            f.write("#\n")
            f.write("# area_info:\n")
            f.write(f"Total_area = {area * 1e-6} <km²>\n")
            f.write("#\n")
            f.write("# crater_diameters\n")
            f.write("crater = {diam, fraction, lon, lat, topo_scale_factor\n")
            for crater in crater_list:
                f.write(
                    f"{crater.measured_diameter * 1e-3}\t{overlap_fraction(crater, region_poly)}\t{crater.measured_location[0]}\t{crater.measured_location[1]}\t 1\n"
                )
            f.write("}\n")

        return

    def from_scc_file(self, input_file: Path | str) -> list[Crater]:
        """
        Import crater data from a Spatial Crater Count file.

        Parameters
        ----------
        input_file : Path | str
            The path to the SCC file containing crater data.

        Returns
        -------
        list[Crater]
            A list of Crater objects imported from the SCC file.
        """
        from craterstats import Spatialcount

        craters = []
        input_file = Path(input_file)
        if not input_file.exists():
            raise FileNotFoundError(f"Input file '{input_file}' does not exist.")
        if input_file.suffix != ".scc":
            raise ValueError(f"Input file '{input_file}' is not a .scc file.")
        scc = Spatialcount(filename=str(input_file))
        for diam, lon, lat in zip(scc.diam, scc.lon, scc.lat, strict=True):
            crater = self.Crater.maker(diameter=diam * 1e3, location=(lon, lat), morphology=self.morphology)
            craters.append(crater)

        return craters

    def from_xarray(self, dataset: xr.Dataset | dict, interval: int | None = None) -> list[Crater]:
        """
        Import crater data from an xarray Dataset.

        Parameters
        ----------
        dataset : xr.Dataset | dict
            The xarray Dataset containing crater data or a dictionary of xarray Datasets keyed by interval number.

        Returns
        -------
        list[Crater]
            A list of Crater objects imported from the xarray Dataset.
        """
        craters = []
        if type(dataset) is dict:
            if interval is None:
                dataset = dataset[-1]
            elif interval in dataset:
                dataset = dataset[interval]
            else:
                return craters
        if "interval" in dataset.coords:
            if interval is None:
                dataset = dataset.isel(interval=-1)
            elif interval in dataset.interval:
                dataset = dataset.sel(interval=interval)
            else:
                return craters
        dataset.load()
        if len(dataset) == 0:
            return craters
        for id in tqdm(dataset.id.data, desc="Converting xarray Dataset to Crater objects", unit="crater", position=0, leave=False):
            crater_data = dataset.sel(id=id).to_dict()["data_vars"]
            crater_data = {k: v["data"] for k, v in crater_data.items()}
            if np.isnan(crater_data["semimajor_axis"]):
                continue
            if "longitude" in crater_data and "latitude" in crater_data:
                crater_data["location"] = (crater_data.pop("longitude"), crater_data.pop("latitude"))
            if "measured_longitude" in crater_data and "measured_latitude" in crater_data:
                crater_data["measured_location"] = (
                    crater_data.pop("measured_longitude"),
                    crater_data.pop("measured_latitude"),
                )
            for k, v in crater_data.items():
                if v is not None and np.any(np.isreal(v)) and np.any(np.isnan(v)):
                    crater_data[k] = None
            crater = self.Crater.maker(**crater_data, morphology=self.morphology, check_redundant_inputs=False)
            craters.append(crater)

        return craters

    def remove_complex_data(self):
        """
        Remove complex data from all observed and emplaced craters to free up memory. This is typically called after the tally step to clear out things like the affect node and face index sets.
        """
        for v in self.observed.values():
            v.remove_complex_data()
        for v in self.emplaced:
            v.remove_complex_data()
        return

    @property
    def surface(self):
        """The Surface object associated with this Counting object."""
        return self._surface

    @surface.setter
    def surface(self, value):
        from cratermaker.components.surface import LocalSurface, Surface

        if not isinstance(value, (Surface | LocalSurface)):
            raise TypeError("surface must be an instance of Surface or LocalSurface")
        self._surface = value

    @property
    def n_layer(self) -> int:
        """Number of layers in the counting model."""
        return _N_LAYER

    @property
    def observed(self) -> dict[int, Crater]:
        """Dictionary of observed craters on the surface keyed to the crater id."""
        return self._observed

    @property
    def emplaced(self) -> list[Crater]:
        """List of craters that have been emplaced in the simulation in the current interval in chronological order."""
        return self._emplaced

    @property
    def n_emplaced(self) -> int:
        """Number of craters that have been emplaced in the simulation in the current interval."""
        return len(self._emplaced)

    @property
    def n_observed(self) -> int:
        """Number of craters that have been observed on the surface in the current interval."""
        return len(self._observed)

    @property
    def Crater(self) -> type[Crater]:
        """The Crater class used for this counting component, which is determined by the morphology component."""
        if self._Crater is None:
            return Crater
        return self._Crater

    @Crater.setter
    def Crater(self, value):
        if value is None:
            return
        if not isinstance(value, type) or not issubclass(value, Crater):
            raise TypeError("Crater must be a subclass of the base Crater class.")
        self._Crater = value

    @property
    def morphology(self) -> Morphology:
        """The morphology component associated with this counting component, which determines the Crater class and the crater properties that are tracked in the simulation."""
        return self._morphology

    @morphology.setter
    def morphology(self, value):
        from cratermaker.components.morphology import Morphology

        if not isinstance(value, Morphology):
            raise TypeError("morphology must be an instance of the Morphology class.")
        self._morphology = value
        if not issubclass(self.Crater, value.Crater):
            self.Crater = value.Crater  # Set the Crater class to the one specified by the morphology component if they don't match
        return


def R_to_CSFD(
    R: Callable[[FloatLike | ArrayLike], FloatLike | ArrayLike],
    D: FloatLike | ArrayLike,
    Dlim: FloatLike = 1e6,
    *args: Any,
) -> FloatLike | ArrayLike:
    """
    Convert R values to cumulative N values for a given D using the R-plot function.

    Parameter
    ----------
    R : R = f(D)
        A function that computes R given D.
    D : FloatLike or ArrayLike
        diameter in units of km.
    Dlim : FloatLike
        Upper limit on the diameter over which to evaluate the integral
    args : Any
        Additional arguments to pass to the R function

    Returns
    -------
    float or ArrayLike
        The cumulative number of craters greater than D in diameter.
    """

    def _R_to_CSFD_scalar(R, D, Dlim, *args):
        # Helper function to integrate the R function
        def integrand(D):
            return R(D, *args) / D**3  # This is dN/dD

        N = 0.0
        D_i = D
        while D_i < Dlim:
            D_next = D_i * np.sqrt(2.0)
            D_mid = (D_i + D_next) / 2  # Mid-point of the bin
            bin_width = D_next - D_i
            R_value = integrand(D_mid)
            N += R_value * bin_width
            D_i = D_next  # Move to the next bin

        return N

    return _R_to_CSFD_scalar(R, D, Dlim, *args) if np.isscalar(D) else np.vectorize(_R_to_CSFD_scalar)(R, D, Dlim, *args)


def csfd_geometric_saturation(diameter: FloatLike | ArrayLike) -> FloatLike | ArrayLike:
    """
    Calculate the cumulative number of craters at geometric saturation for a given diameter.

    We use the definition of geomatric saturation from Melosh (1989) [#]_.

    Parameter
    ----------
    diameter : FloatLike or ArrayLike
        The diameter(s) for which to calculate the cumulative number of craters at geometric saturation.

    Returns
    -------
    FloatLike or ArrayLike
        The cumulative number of craters at geometric saturation for the given diameter(s).

    References
    ----------
    .. [#] Melosh, H.J., 1989. Impact cratering: A geologic process. Oxford University Press, New York, New York.

    """
    return 1.54 * diameter ** (-2)


def csfd_equilibrium(diameter: FloatLike | ArrayLike, f_geometric=0.0218) -> FloatLike | ArrayLike:
    """
    Calculate the cumulative number of craters at equilibrium for a given diameter.

    Parameter
    ----------
    diameter : FloatLike or ArrayLike
        The diameter(s) for which to calculate the cumulative number of craters at equilibrium.
    f_geometric : float, optional
        The fraction of geometric saturation at which equilibrium occurs. The default value is 0.0218, which is the value used in Minton et al. (2019) [#]_.


    Returns
    -------
    FloatLike or ArrayLike
        The cumulative number of craters at equilibrium for the given diameter(s).

    References
    ----------
    .. [#] Minton, D.A., Fassett, C.I., Hirabayashi, M., Howl, B.A., Richardson, J.E., (2019). The equilibrium size-frequency distribution of small craters reveals the effects of distal ejecta on lunar landscape morphology. Icarus 326, 63-87. https://doi.org/10.1016/j.icarus.2019.02.021

    """
    return f_geometric * csfd_geometric_saturation(diameter)


import_components(__name__, __path__)



from typing import Any
from warnings import warn

from numpy.random import Generator

from cratermaker.components.projectile import Projectile
from cratermaker.components.target import Target
from cratermaker.constants import FloatLike


@Projectile.register("comets")
class CometProjectiles(Projectile):
    _catalogue = {
        "Mercury": 18400.0,
        "Venus": 26300.0,
        "Earth": 17670.0,
        "Moon": 17670.0,
        "Mars": 14210.0,
        "Metis": 59000.0,
        "Amalthea": 50000.0,
        "Thebe": 45000.0,
        "Io": 32000.0,
        "Europa": 26000.0,
        "Ganymede": 20000.0,
        "Callisto": 15000.0,
        "Himalia": 6100.0,
        "Prometheus": 32000.0,
        "Pandora": 31000.0,
        "Epimetheus": 30000.0,
        "Janus": 30000.0,
        "Mimas": 27000.0,
        "Enceladus": 24000.0,
        "Tethys": 21000.0,
        "Telesto": 21000.0,
        "Calypso": 21000.0,
        "Dione": 19000.0,
        "Helene": 19000.0,
        "Rhea": 16000.0,
        "Titan": 10500.0,
        "Hyperion": 9400.0,
        "Iapetus": 6100.0,
        "Phoebe": 3200.0,
        "Cordelia": 20000.0,
        "Ophelia": 19000.0,
        "Bianca": 18000.0,
        "Cressida": 18000.0,
        "Desdemona": 18000.0,
        "Juliet": 18000.0,
        "Portia": 18000.0,
        "Rosalind": 17000.0,
        "Belinda": 16000.0,
        "Puck": 15000.0,
        "Miranda": 12500.0,
        "Ariel": 10300.0,
        "Umbriel": 8700.0,
        "Titania": 6800.0,
        "Oberon": 5900.0,
        "Naiad": 22000.0,
        "Thalassa": 22000.0,
        "Despina": 21000.0,
        "Galatea": 20000.0,
        "Larissa": 18000.0,
        "Proteus": 14000.0,
        "Triton": 8200.0,
        "Nereid": 2800.0,
        "Pluto": 1900.0,
        "Charon": 1800.0,
        "KBO": 1800.0,
    }

    def __init__(
        self,
        sample: bool = True,
        density: FloatLike | None = None,
        angle: FloatLike | None = None,
        direction: FloatLike | None = None,
        location: tuple[FloatLike, FloatLike] | None = None,
        target: Target | str | None = None,
        rng: Generator | None = None,
        rng_seed: int | None = None,
        rng_state: dict | None = None,
        **kwargs: Any,
    ):
        """
        An operations class for computing the projectile properties of a comet source population.

        Parameters
        ----------
        sample : bool, default True
            Flag that determines whether to sample impact velocities, angles, and directions from distributions. If set to False, the projectile velocity will be the mean velocity for the given target body.
        density : float
            The density of the projectile in kg/m³. Default is 500.0 kg/m³.
        angle : float, optional
            The impact angle in degrees. Default is 90.0 degrees (vertical impact) if `sample` is False. If `sample` is True, this value is ignored.
        direction : float | None
            The impact direction in degrees. Default is 0.0 degrees (due North) if `sample` is False. If `sample` is True, this value is ignored.`
        location : tuple[float, float] | None
            The impact location as a tuple of (longitude, latitude) in degrees. Default is (0.0, 0.0) (the equator and prime meridian) if `sample` is False. If `sample` is True, this value is ignored.
        target : Target or str.
            The name of the target body for the impact. Default is "Moon"
        rng : Generator | None
            |rng|
        rng_seed : int | None
            The random rng_seed for the simulation if rng is not provided. If None, a random rng_seed is used.
        rng_state : dict, optional
            |rng_state|

        Notes
        -----
        The comets are assumed to be Jupiter-family Comets. The mean impact velocities for outer solar system bodies come from Table 1 of Zahnle et al. [#]_, and for inner solar system bodies, Table 2 of Borin et al. [#]_.

        References
        ----------
        .. [#] Zahnle, K., Schenk, P., Levison, H., Dones, L., 2003. Cratering rates in the outer Solar System. Icarus 163, 263-289. https://doi.org/10.1016/S0019-1035(03)00048-4
        .. [#] Borin, P., Cremonese, G., Marzari, F., Lucchetti, A., 2017. Asteroidal and cometary dust flux in the inner solar system. A&A 605, A94. https://doi.org/10.1051/0004-6361/201730617


        **kwargs : Any
            |kwargs|
        """
        density = 500.0 if density is None else density
        kwargs.pop("mean_velocity", None)
        kwargs.pop("velocity", None)
        self._target = Target.maker(target, **kwargs)
        mean_velocity = self._set_mean_velocity()
        super().__init__(
            sample=sample,
            mean_velocity=mean_velocity,
            velocity=mean_velocity,
            angle=angle,
            direction=direction,
            location=location,
            density=density,
            target=target,
            rng=rng,
            rng_seed=rng_seed,
            rng_state=rng_state,
            **kwargs,
        )

    def __str__(self) -> str:
        base = super().__str__()
        return f"{base}\nTarget: {self.target.name}\n"

    def _set_mean_velocity(self):
        """
        Sets the mean velocity of the projectile in m/s based on the target body.

        Returns
        -------
        float
        """
        if self.target.name in self.__class__._catalogue:
            pmv = float(self.__class__._catalogue[self.target.name])
        else:
            warn(
                f"Target {self.target.name} not found in known targets. Known targets include {list(self.__class__._catalogue.keys())}. Defaulting to KBO.",
                stacklevel=2,
            )
            pmv = float(self.__class__._catalogue["KBO"])
        return pmv



import numpy as np

# Custom type aliases for better readability
FloatLike = float | int | np.number
PairOfFloats = tuple[float, float] | list[float] | np.ndarray

# Default filenames and paths
_CONFIG_FILE_NAME = "cratermaker.yaml"
_COMPONENT_NAMES = [
    "counting",
    "target",
    "scaling",
    "production",
    "morphology",
    "projectile",
    "surface",
]
# This is a factor used to determine the smallest length scale in the grid
_SMALLFAC = 1.0e-5
_VSMALL = 10 * np.finfo(np.float64).tiny
_LOGVSMALL = np.log10(_VSMALL)

# Map of OGR drivers to file extensions
VECTOR_DRIVER_TO_EXTENSION_MAP = {
    "PCIDSK": "pix",
    "PDS4": "xml",
    "PDF": "pdf",
    "MBTILES": "mbtiles",
    "ESRI SHAPEFILE": "shp",
    "MAPINFO FILE": "tab",
    "S57": "000",
    "DGN": "dgn",
    "CSV": "csv",
    "GML": "gml",
    "GPX": "gpx",
    "KML": "kml",
    "GEOJSON": "json",
    "GEOJSONSEQ": "geojsonl",
    "OGR_GMT": "gmt",
    "GPKG": "gpkg",
    "SQLite": "sqlite",
    "WASP": "map",
    "OPENFILEGDB": "gdb",
    "DXF": "dxf",
    "FLATGEOBUF": "fgb",
    "PGDUMP": "sql",
    "GPSBABEL": "mps",
    "ODS": "ods",
    "XLSX": "xlsx",
    "JML": "jml",
    "VDV": "txt",
    "MVT": "mvt",
    "PMTiles": "pmtiles",
    "JSONFG": "json",
    "MIRAMONVECTOR": "pol",
}


# Optional: controlled public API
__all__ = []
